from __future__ import annotations

import io
import json
import logging
from copy import copy
from pathlib import Path
from typing import TYPE_CHECKING, Iterable
from warnings import catch_warnings, filterwarnings

import dask.array as da
import fsspec
import numpy as np
import zarr
import zarr.storage
from natsort import natsorted
from numpy.typing import ArrayLike
from tifffile import TiffFile, ZarrTiffStore
from xarray import DataArray

from iohub.mm_fov import MicroManagerFOV, MicroManagerFOVMapping

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath


__all__ = ["MMOmeTiffFOV", "MMStack"]
_logger = logging.getLogger(__name__)


def _normalize_mm_pos_key(key: str | int) -> int:
    try:
        return int(key)
    except TypeError:
        raise TypeError("Micro-Manager position keys must be integers.")


def _tiff_to_fsspec_store(
    zarr_tiff_store: ZarrTiffStore, root_uri: str
) -> zarr.storage.FsspecStore:
    """Bridge tifffile (zarr-python v2 interface) with zarr-python v3.

    Parameters
    ----------
    zarr_tiff_store : ZarrTiffStore
        Zarr (v2) wrapper for a TIFF series
    root_uri : str
        `file://` URI to the directory containing the TIFF files

    Returns
    -------
    zarr.storage.FsspecStore
        Zarr (v3) wrapper for a TIFF series
    """
    spec_container = io.StringIO()
    zarr_tiff_store.write_fsspec(spec_container, url=root_uri)
    fs, _ = fsspec.url_to_fs(
        "reference://",
        fo=json.loads(spec_container.getvalue()),
        target_protocol="file",
        asynchronous=True,
    )
    return zarr.storage.FsspecStore(fs=fs)


def find_first_ome_tiff_in_mmstack(data_path: Path) -> Path:
    if data_path.is_file():
        if "ome.tif" in data_path.name:
            return data_path
        else:
            raise ValueError("{data_path} is not a OME-TIFF file.")
    elif data_path.is_dir():
        files = data_path.glob("*.ome.tif")
        try:
            return next(files)
        except StopIteration:
            raise FileNotFoundError(
                f"Path {data_path} contains no OME-TIFF files."
            )
    raise FileNotFoundError(f"Path {data_path} does not exist.")


class MMOmeTiffFOV(MicroManagerFOV):
    def __init__(self, parent: MMStack, key: str) -> None:
        super().__init__(parent, key)
        self._xdata = parent.xdata[key]

    @property
    def axes_names(self) -> list[str]:
        return list(self._xdata.dims)

    @property
    def shape(self) -> tuple[int, int, int, int, int]:
        return self._xdata.shape

    @property
    def dtype(self) -> np.dtype:
        return self._xdata.dtype

    def __getitem__(self, key: int | slice | tuple[int | slice]) -> ArrayLike:
        return self._xdata[key]

    @property
    def xdata(self) -> DataArray:
        return self._xdata

    def frame_metadata(self, t: int, c: int, z: int) -> dict | None:
        """Read image plane metadata from the OME-TIFF file."""
        return self.parent.read_image_metadata(self._position, t, c, z)


class MMStack(MicroManagerFOVMapping):
    """Micro-Manager multi-file OME-TIFF (MMStack) reader.

    Parameters
    ----------
    data_path : StrOrBytesPath
        Path to the directory containing OME-TIFF files
        or the path to the first OME-TIFF file in the series
    """

    def __init__(self, data_path: StrOrBytesPath):
        super().__init__()
        data_path = Path(data_path)
        first_file = find_first_ome_tiff_in_mmstack(data_path)
        self._root = first_file.parent
        self.dirname = self._root.name
        self._first_tif = TiffFile(first_file, is_mmstack=True)
        _logger.debug(f"Parsing {first_file} as MMStack.")
        with catch_warnings():
            # The IJMetadata tag (50839) is sometimes not written
            # See https://micro-manager.org/Micro-Manager_File_Formats
            filterwarnings("ignore", message=r".*50839.*", module="tifffile")
            self._parse_data()
        self._store = None

    def _parse_data(self):
        series = self._first_tif.series[0]
        raw_dims = dict(
            (axis, size)
            for axis, size in zip(series.get_axes(), series.get_shape())
        )
        axes = ("R", "T", "C", "Z", "Y", "X")
        dims = dict((ax, raw_dims.get(ax, 1)) for ax in axes)
        _logger.debug(f"Got dataset dimensions from tifffile: {dims}.")
        (
            self.positions,
            self.frames,
            self.channels,
            self.slices,
            self.height,
            self.width,
        ) = dims.values()
        self._set_mm_meta(self._first_tif.micromanager_metadata)
        zarr_tiff_store = series.aszarr(multiscales=True)
        self._store = _tiff_to_fsspec_store(
            zarr_tiff_store, root_uri=self._root.as_uri()
        )
        _logger.debug(f"Opened {self._store}.")
        data = da.from_zarr(zarr.open(self._store, mode="r")["0"])
        self.dtype = data.dtype
        img = DataArray(data, dims=raw_dims, name=self.dirname)
        xarr = img.expand_dims(
            [ax for ax in axes if ax not in img.dims]
        ).transpose(*axes)
        if self.channels > len(self.channel_names):
            for c in range(self.channels):
                if c >= len(self.channel_names):
                    self.channel_names.append(f"Channel{c}")
            _logger.warning(
                "Number of channel names in the metadata is "
                "smaller than the number of channels. "
                f"Completing with fallback names: {self.channel_names}."
            )
        # not all positions in the position list may have been acquired
        xarr = xarr[: self.positions]
        xarr.coords["C"] = self.channel_names
        xset = xarr.to_dataset(dim="R")
        self._xdata = xset
        self._infer_image_meta()

    @property
    def xdata(self):
        return self._xdata

    def __len__(self) -> int:
        return len(self.xdata.keys())

    def __getitem__(self, key: str | int) -> MMOmeTiffFOV:
        key = _normalize_mm_pos_key(key)
        return MMOmeTiffFOV(self, key)

    def __setitem__(self, key, value) -> None:
        raise PermissionError("MMStack is read-only.")

    def __delitem__(self, key, value) -> None:
        raise PermissionError("MMStack is read-only.")

    def __contains__(self, key: str | int) -> bool:
        key = _normalize_mm_pos_key(key)
        return key in self.xdata

    def __iter__(self) -> Iterable[tuple[str, MMOmeTiffFOV]]:
        for key in self.xdata:
            yield str(key), self[key]

    def __enter__(self) -> MMStack:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()

    def close(self) -> None:
        """Close file handles"""
        self._first_tif.close()

    def _set_mm_meta(self, mm_meta: dict) -> None:
        """Assign image metadata from summary metadata."""
        self._mm_meta = mm_meta
        self.channel_names = []
        mm_version = self._mm_meta["Summary"]["MicroManagerVersion"]
        if "beta" in mm_version:
            if self._mm_meta["Summary"]["Positions"] > 1:
                self._stage_positions = []

                for p in range(
                    len(self._mm_meta["Summary"]["StagePositions"])
                ):
                    pos = self._simplify_stage_position_beta(
                        self._mm_meta["Summary"]["StagePositions"][p]
                    )
                    self._stage_positions.append(pos)

            # MM beta versions sometimes don't have 'ChNames',
            # so I'm wrapping in a try-except and setting the
            # channel names to empty strings if it fails.
            try:
                for ch in self._mm_meta["Summary"]["ChNames"]:
                    self.channel_names.append(ch)
            except Exception:
                self.channel_names = self._mm_meta["Summary"]["Channels"] * [
                    ""
                ]  # empty strings

        elif mm_version == "1.4.22":
            for ch in self._mm_meta["Summary"].get("ChNames", []):
                self.channel_names.append(ch)

        # Parsing of data acquired with the OpenCell
        # acquisition script on the Dragonfly miroscope
        elif (
            mm_version == "2.0.1 20220920"
            and self._mm_meta["Summary"].get("Prefix", None) == "raw_data"
        ):
            files = natsorted(self.root.glob("*.ome.tif"))
            self.positions = len(files)  # not all positions are saved

            if self._mm_meta["Summary"]["Positions"] > 1:
                self._stage_positions = [None] * self.positions

                for p_idx, file_name in enumerate(files):
                    site_idx = int(str(file_name).split("_")[-1].split("-")[0])
                    pos = self._simplify_stage_position(
                        self._mm_meta["Summary"]["StagePositions"][site_idx]
                    )
                    self._stage_positions[p_idx] = pos

            for ch in self._mm_meta["Summary"]["ChNames"]:
                self.channel_names.append(ch)

        else:
            if self._mm_meta["Summary"].get("Positions", 1) > 1:
                self._stage_positions = []

                for p in range(self._mm_meta["Summary"]["Positions"]):
                    pos = self._simplify_stage_position(
                        self._mm_meta["Summary"]["StagePositions"][p]
                    )
                    self._stage_positions.append(pos)

            for ch in self._mm_meta["Summary"].get("ChNames", []):
                self.channel_names.append(ch)
        z_step_size = float(self._mm_meta["Summary"].get("z-step_um", 1.0))
        if z_step_size == 0:
            if self.slices == 1:
                z_step_size = 1.0
            else:
                _logger.warning(
                    f"Z-step size is {z_step_size} um in the metadata, "
                    "Using 1.0 um instead."
                )
        self._z_step_size = z_step_size
        self.height = self._mm_meta["Summary"]["Height"]
        self.width = self._mm_meta["Summary"]["Width"]
        self._t_scale = (
            float(self._mm_meta["Summary"].get("Interval_ms", 1e3)) / 1e3
        )

    def _simplify_stage_position(self, stage_pos: dict):
        """
        flattens the nested dictionary structure of stage_pos
        and removes superfluous keys

        Parameters
        ----------
        stage_pos:      (dict)
            dictionary containing a single position's device info

        Returns
        -------
        out:            (dict)
            flattened dictionary
        """

        out = copy(stage_pos)
        out.pop("DevicePositions")
        for dev_pos in stage_pos["DevicePositions"]:
            out.update({dev_pos["Device"]: dev_pos["Position_um"]})
        return out

    def _simplify_stage_position_beta(self, stage_pos: dict) -> dict:
        """
        flattens the nested dictionary structure of stage_pos
        and removes superfluous keys
        for MM2.0 Beta versions

        Parameters
        ----------
        stage_pos:      (dict)
            dictionary containing a single position's device info

        Returns
        -------
        new_dict:       (dict)
            flattened dictionary

        """

        new_dict = {}
        new_dict["Label"] = stage_pos["label"]
        new_dict["GridRow"] = stage_pos["gridRow"]
        new_dict["GridCol"] = stage_pos["gridCol"]

        for sub in stage_pos["subpositions"]:
            values = []
            for field in ["x", "y", "z"]:
                if sub[field] != 0:
                    values.append(sub[field])
            if len(values) == 1:
                new_dict[sub["stageName"]] = values[0]
            else:
                new_dict[sub["stageName"]] = values

        return new_dict

    def read_image_metadata(
        self, p: int, t: int, c: int, z: int
    ) -> dict | None:
        """Read image plane metadata from the OME-TIFF file."""
        multi_index = (p, t, c, z)
        tif_shape = (self.positions, self.frames, self.channels, self.slices)
        idx = np.ravel_multi_index(multi_index, tif_shape)
        return self._read_frame_metadata(idx)

    def _read_frame_metadata(self, idx: int) -> dict | None:
        # `TiffPageSeries` is not a collection of `TiffPage` objects
        # but a mixture of `TiffPage` and `TiffFrame` objects
        # https://github.com/cgohlke/tifffile/issues/179
        with catch_warnings():
            filterwarnings(
                "ignore", message=r".*from closed file.*", module="tifffile"
            )
            try:
                # virtual frames
                page = self._first_tif.pages[idx]
            except IndexError:
                page = self._first_tif.series[0].pages[idx]
            if page:
                try:
                    page = page.aspage()
                except ValueError:
                    _logger.warning("Cannot read tags from virtual frame.")
                    return None
            else:
                # invalid page
                _logger.warning(f"Page {idx} is not present in the dataset.")
                return None
            try:
                return page.tags["MicroManagerMetadata"].value
            except KeyError:
                _logger.warning("The Micro-Manager metadata tag is not found.")
                return None

    def _infer_image_meta(self) -> None:
        """
        Infer data type and pixel size from the first image plane metadata.
        """
        _logger.debug("Inferring image metadata.")
        metadata = self._read_frame_metadata(0)
        if metadata is not None:
            try:
                self._xy_pixel_size = float(metadata["PixelSizeUm"])
                if self._xy_pixel_size > 0:
                    return
            except Exception:
                _logger.warning(
                    "Micro-Manager image plane metadata cannot be loaded."
                )
        _logger.warning(
            "XY pixel size cannot be determined, defaulting to 1.0 um."
        )
        self._xy_pixel_size = 1.0

    @property
    def zyx_scale(self) -> tuple[float, float, float]:
        return (
            self._z_step_size,
            self._xy_pixel_size,
            self._xy_pixel_size,
        )

    @property
    def t_scale(self) -> float:
        return self._t_scale



"""
iohub
=====

N-dimensional bioimaging data I/O with OME metadata in Python
"""

import logging
import os

from iohub.ngff import open_ome_zarr
from iohub.reader import read_images

__all__ = ["open_ome_zarr", "read_images"]


def _configure_logging():
    """Configure logging for iohub."""
    level = os.environ.get("IOHUB_LOG_LEVEL", logging.INFO)
    if str(level).isdigit():
        level = int(level)
    iohub_logger = logging.getLogger(__name__)
    iohub_logger.setLevel(level)
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(
        logging.Formatter(fmt="%(levelname)s:%(name)s:%(message)s")
    )
    iohub_logger.addHandler(stream_handler)


_configure_logging()



"""
Writing a Large Array
=====================

This script shows how to store a larger-than-RAM array to a OME-Zarr dataset.
The same method works for HCS datasets too.
"""

# %%
# .. warning:: Executing this example will write a large file.
#     Modify the store path manually.
import numpy as np
from tqdm import tqdm

from iohub.ngff import open_ome_zarr

store_path = ""

# %%
# Shape and data type of the large array
# this array is about 10 GB, each time point is about 100 MB.
# It may not be actually larger than RAM but enough for demo.
# Monitor the memory usage of Python when the following runs
# and it should take significantly less than 10 GB.
shape = (100, 2, 25, 1024, 1024)
dtype = np.uint16


# %%
# .. note:: This will not run if the Zarr store path is not set above
#
# Store this array by looping through the time points
if store_path:
    with open_ome_zarr(
        store_path, layout="fov", mode="w-", channel_names=["DAPI", "GFP"]
    ) as dataset:
        img = dataset.create_zeros(
            name="0",
            shape=shape,
            dtype=dtype,
            chunks=(1, 1, 1, 1024, 1024),  # chunk by XY planes
        )
        for t in tqdm(range(shape[0])):
            # write 4D image data for the time point
            img[t] = np.ones(shape[1:]) * t
        dataset.print_tree()



import csv
import os
import shutil
from pathlib import Path

import fsspec
import numpy as np
import pytest
from wget import download


def _download_ndtiff_v3_labeled_positions(test_data: Path) -> None:
    ghfs = fsspec.filesystem(
        "github",
        org="micro-manager",
        repo="NDTiffStorage",
        username=os.environ.get("GITHUB_ACTOR"),
        token=os.environ.get("GITHUB_TOKEN"),
    )
    v3_lp = test_data / "ndtiff_v3_labeled_positions"
    Path.mkdir(v3_lp)
    ghfs.get(
        ghfs.ls("test_data/v3/labeled_positions_1"),
        str(v3_lp),
        recursive=True,
    )


def download_data():
    """Download test datasets."""
    test_data = Path.cwd() / ".pytest_temp" / "test_data"
    if not test_data.is_dir():
        Path.mkdir(test_data, parents=True)
        print("\nsetting up temp folder")

    # Zenodo URL
    custom_url = (
        "https://zenodo.org/record/6983916/files/waveOrder_test_data.zip"
    )
    # Reference v0.4 HCS dataset from OME
    # See the last line of
    # https://github.com/ome/ngff/issues/140#issuecomment-1309972511
    ome_hcs_url = "https://zenodo.org/record/8091756/files/20200812-CardiomyocyteDifferentiation14-Cycle1.zarr.zip"  # noqa

    # download files to temp folder
    if not any(test_data.iterdir()):
        print("Downloading test files...")
        for url in (custom_url, ome_hcs_url):
            output = test_data / Path(url).name
            download(url, out=str(output))
            shutil.unpack_archive(output, extract_dir=test_data)
        _download_ndtiff_v3_labeled_positions(test_data)
    return test_data


def subdirs(parent: Path, name: str) -> list[Path]:
    return [d for d in (parent / name).iterdir() if d.is_dir()]


test_datasets = download_data()


mm2gamma_ome_tiffs = subdirs(test_datasets, "MM20_ome-tiffs")


mm2gamma_ome_tiffs_hcs = [p for p in mm2gamma_ome_tiffs if "4p" in p.name]


# This is a dataset with 11 timepoints
# The MDA definition at start of the experiment specifies 20 timepoints
mm2gamma_ome_tiffs_incomplete = (
    test_datasets
    / "MM20_ometiff_incomplete"
    / "mm2.0-20201209_20t_5z_3c_512k_incomplete_1"
)


mm2gamma_singlepage_tiffs = subdirs(test_datasets, "MM20_singlepage-tiffs")


# This is a dataset with 11 timepoints
# The MDA definition at start of the experiment specifies 20 timepoints
mm2gamma_singlepage_tiffs_incomplete = (
    test_datasets
    / "MM20_singlepage_incomplete"
    / "mm2.0-20201209_20t_5z_3c_512k_incomplete_1 2"
)


mm1422_ome_tiffs = subdirs(test_datasets, "MM1422_ome-tiffs")


mm1422_singlepage_tiffs = subdirs(test_datasets, "MM1422_singlepage-tiffs")


mm2gamma_zarr_v01 = (
    test_datasets / "MM20_zarr" / "mm2.0-20201209_4p_2t_5z_1c_512k_1.zarr"
)


hcs_ref = test_datasets / "20200812-CardiomyocyteDifferentiation14-Cycle1.zarr"


ndtiff_v2_datasets = subdirs(test_datasets, "MM20_pycromanager")


ndtiff_v2_ptcz = (
    test_datasets
    / "MM20_pycromanager"
    / "mm2.0-20210713_pm0.13.2_2p_3t_2c_7z_1"
)


ndtiff_v3_labeled_positions = test_datasets / "ndtiff_v3_labeled_positions"


@pytest.fixture
def csv_data_file_1(tmpdir):
    test_csv_1 = tmpdir / "well_names_1.csv"
    csv_data_1 = [
        ["B/03", "D/4"],
    ]
    with open(test_csv_1, mode="w", newline="") as csvfile:
        writer = csv.writer(csvfile)
        writer.writerows(csv_data_1)
    return test_csv_1


@pytest.fixture
def csv_data_file_2(tmpdir):
    test_csv_2 = tmpdir / "well_names_2.csv"
    csv_data_2 = [
        ["D/4", "B/03"],
    ]
    with open(test_csv_2, mode="w", newline="") as csvfile:
        writer = csv.writer(csvfile)
        writer.writerows(csv_data_2)
    return test_csv_2


@pytest.fixture
def empty_ome_zarr_hcs_v05(tmpdir) -> tuple[Path, tuple[tuple[str, ...], ...]]:
    """Create an empty HCS OME-Zarr v0.5 dataset."""
    example_json_dir = Path(__file__).parent / "ngff" / "static_data" / "v05"
    empty_zarr = tmpdir / "v05.hcs.ome.zarr"
    empty_zarr.mkdir()
    TARGET_FILENAME = "zarr.json"
    shutil.copy(example_json_dir / "plate.json", empty_zarr / TARGET_FILENAME)
    ROWS = ("A", "B")
    COLS = ("1", "2", "3")
    FOVS = ("0", "1", "2", "3")
    RESOLUTIONS = ("0", "1", "2")
    for row in ROWS:
        row_dir = empty_zarr / row
        row_dir.mkdir()
        shutil.copy(example_json_dir / "row.json", row_dir / TARGET_FILENAME)
        for col in COLS:
            col_dir = row_dir / col
            col_dir.mkdir()
            shutil.copy(
                example_json_dir / "well.json", col_dir / TARGET_FILENAME
            )
            for fov in FOVS:
                fov_dir = col_dir / fov
                fov_dir.mkdir()
                shutil.copy(
                    example_json_dir / "image.json", fov_dir / TARGET_FILENAME
                )
                for res in RESOLUTIONS:
                    res_dir = fov_dir / res
                    res_dir.mkdir()
                    shutil.copy(
                        example_json_dir / "array.json",
                        res_dir / TARGET_FILENAME,
                    )
    return empty_zarr, (ROWS, COLS, FOVS, RESOLUTIONS)


@pytest.fixture()
def aqz_ome_zarr_05(tmpdir):
    pytest.importorskip("acquire_zarr")
    import acquire_zarr as aqz

    store_path = tmpdir / "ome_zarr_v0.5.zarr"

    settings = aqz.StreamSettings(
        arrays=[
            aqz.ArraySettings(
                data_type=np.uint16,
                compression=aqz.CompressionSettings(
                    codec=aqz.CompressionCodec.BLOSC_LZ4,
                    compressor=aqz.Compressor.BLOSC1,
                    level=1,
                    shuffle=0,
                ),
                dimensions=[
                    aqz.Dimension(
                        name="t",
                        kind=aqz.DimensionType.TIME,
                        array_size_px=0,
                        chunk_size_px=16,
                        shard_size_chunks=1,
                    ),
                    aqz.Dimension(
                        name="c",
                        kind=aqz.DimensionType.CHANNEL,
                        array_size_px=4,
                        chunk_size_px=1,
                        shard_size_chunks=1,
                    ),
                    aqz.Dimension(
                        name="z",
                        kind=aqz.DimensionType.SPACE,
                        array_size_px=10,
                        chunk_size_px=10,
                        shard_size_chunks=1,
                    ),
                    aqz.Dimension(
                        name="y",
                        kind=aqz.DimensionType.SPACE,
                        array_size_px=48,
                        chunk_size_px=16,
                        shard_size_chunks=3,
                    ),
                    aqz.Dimension(
                        name="x",
                        kind=aqz.DimensionType.SPACE,
                        array_size_px=64,
                        chunk_size_px=16,
                        shard_size_chunks=2,
                    ),
                ],
                downsampling_method=aqz.DownsamplingMethod.MEAN,
            )
        ],
        store_path=str(store_path),
        version=aqz.ZarrVersion.V3,
        max_threads=1,
    )

    stream = aqz.ZarrStream(settings)
    data = np.random.randint(
        0, 2**16 - 1, (32, 4, 10, 48, 64), dtype=np.uint16
    )
    stream.append(data)
    del stream

    return store_path



from pathlib import Path

import numpy as np
import pytest

from iohub.ngff.nodes import (
    Position,
    TransformationMeta,
    _pad_shape,
    open_ome_zarr,
)


def _mock_fov(
    tmp_path: Path,
    shape: tuple[int, ...],
    scale: tuple[float, float, float],
) -> Position:
    ds_path = tmp_path / "ds.zarr"
    channels = [str(i) for i in range(shape[1])]

    fov = open_ome_zarr(
        ds_path,
        layout="fov",
        mode="a",
        channel_names=channels,
        axes=Position._DEFAULT_AXES[-len(shape) :],
    )

    transform = [
        TransformationMeta(
            type="scale",
            scale=_pad_shape(scale, len(shape)),
        )
    ]

    if "0" not in fov:
        fov.create_zeros(
            "0",
            shape=shape,
            dtype=np.uint16,
            chunks=_pad_shape(shape[-2:], len(shape)),
            transform=transform,
        )

    return fov


@pytest.mark.skip(reason="zarr-python / ome_zarr incompatibility")
@pytest.mark.parametrize("ndim", [2, 5])
def test_pyramid(tmp_path: Path, ndim: int) -> None:
    from ome_zarr.io import parse_url
    from ome_zarr.reader import Multiscales, Reader

    # not all shapes not divisible by 2
    shape = (2, 2, 67, 115, 128)[-ndim:]
    scale = (2, 0.5, 0.5)[-min(3, ndim) :]
    levels = 4

    fov = _mock_fov(tmp_path, shape, scale)

    fov.initialize_pyramid(levels=levels)

    assert len(fov.array_keys()) == levels

    for level in range(levels):
        assert str(level) in fov.array_keys()

        level_shape = np.asarray(fov[str(level)].shape)
        ratio = np.ceil(shape / level_shape).astype(int)

        assert np.all(ratio[:-3] == 1)  # time and channel aren't scaled
        assert np.all(ratio[-3:] == 2**level)

        level_scale = np.asarray(
            fov.metadata.multiscales[0]
            .datasets[level]
            .coordinate_transformations[0]
            .scale
        )
        assert np.all(level_scale[:-3] == 1)
        assert np.allclose(level_scale[-3:] / scale, 2**level)

        assert fov.metadata.multiscales[0].datasets[level].path == str(level)

    reader = Reader(parse_url(tmp_path / "ds.zarr"))
    for node in reader():
        assert any(isinstance(spec, Multiscales) for spec in node.specs)



from pathlib import Path

import numpy as np
import pytest

from iohub.clearcontrol import (
    ArrayIndex,
    ClearControlFOV,
    _array_to_blosc_buffer,
    blosc_buffer_to_array,
    create_mock_clear_control_dataset,
)


@pytest.fixture
def mock_clear_control_dataset_path(tmp_path: Path) -> Path:
    ds_path = tmp_path / "dataset.cc"
    create_mock_clear_control_dataset(path=ds_path)
    return ds_path


def test_blosc_buffer(tmp_path: Path) -> None:
    buffer_path = tmp_path / "buffer.blc"
    in_array = np.random.randint(0, 5_000, size=(32, 32))

    _array_to_blosc_buffer(in_array, buffer_path)
    out_array = blosc_buffer_to_array(
        buffer_path, in_array.shape, in_array.dtype
    )

    assert np.allclose(in_array, out_array)


@pytest.mark.parametrize(
    "key",
    [
        1,
        (slice(None), 1),
        (0, [1, 2]),
        (-1, np.asarray([0, 3])),
        (slice(1), -2),
        (np.asarray(0),),
        (0, 0, slice(32)),
    ],
)
def test_CCFOV_indexing(
    mock_clear_control_dataset_path: Path,
    key: ArrayIndex,
) -> None:
    cc = ClearControlFOV(mock_clear_control_dataset_path)

    # copy of whole array
    similar_arr = cc[slice(None)]

    # checking if indexing works as in numpy
    assert np.array_equal(similar_arr[key], cc[key])


def test_CCFOV_metadata(
    mock_clear_control_dataset_path: Path,
) -> None:
    cc = ClearControlFOV(mock_clear_control_dataset_path)
    expected_metadata = {
        "voxel_size_z": 1.0,
        "voxel_size_y": 0.25,
        "voxel_size_x": 0.25,
        "acquisition_type": "NA",
        "time_delta": 45.0,
    }
    metadata = cc.metadata()
    assert metadata == expected_metadata


def test_CCFOV_scales(
    mock_clear_control_dataset_path: Path,
) -> None:
    cc = ClearControlFOV(mock_clear_control_dataset_path)
    zyx_scale = (1.0, 0.25, 0.25)
    time_delta = 45.0
    assert zyx_scale == cc.zyx_scale
    assert time_delta == cc.t_scale


def test_CCFOV_cache(
    mock_clear_control_dataset_path: Path,
) -> None:
    cc = ClearControlFOV(mock_clear_control_dataset_path, cache=True)

    volume = cc[0, 0]
    assert np.array_equal(cc._cache_array, volume)

    cc[0, 0, 0]
    assert np.array_equal(cc._cache_array, volume)
    assert cc._cache_key == (0, 0)

    array = cc[1]
    assert np.array_equal(cc._cache_array, array)

    new_array = cc[1]
    assert id(new_array) == id(array)  # same reference, so cache worked

    cc.cache = False
    assert cc._cache_key is None
    assert cc._cache_array is None



"""
Viewing Clear Control
=====================

Example of opening a Clear Control dataset and viewing with napari.

Usage:

    $ python view_clearcontrol_dataset.py <OPTIONAL Clear Control dataset>

If the dataset path is not provided,
it creates a mock dataset of random integers.
"""

# %%
# Setup

import sys
import tempfile
import time

from iohub.clearcontrol import (
    ClearControlFOV,
    create_mock_clear_control_dataset,
)

# %%
# Parse optional Clear Control dataset path.
# Mock dataset is created if dataset path is not provided.

if len(sys.argv) < 2:
    print("Loading mock random noise dataset ...")
    path = f"{tempfile.gettempdir()}/dataset.cc"
    create_mock_clear_control_dataset(path)
else:
    path = sys.argv[1]

# %%
# Open Clear Control dataset.

s = time.time()
cc = ClearControlFOV(path, cache=True)
print("init time (secs)", time.time() - s)

# %%
# Time load time of a single volume.

s = time.time()
cc[0, 0]
print("single volume load time (secs)", time.time() - s)


# %%
# Load dataset using napari
if __name__ == "__main__":
    try:
        import napari

        s = time.time()
        napari.view_image(cc)
        print("napari load time (secs)", time.time() - s)

        napari.run()

    except ModuleNotFoundError:
        pass



import logging

import iohub  # noqa: F401


def test_external_logging_level(caplog):
    with caplog.at_level(logging.DEBUG):
        iohub_logger = logging.getLogger("iohub")
        assert iohub_logger.getEffectiveLevel() == logging.INFO
        other_logger = logging.getLogger(__name__)
        assert other_logger.getEffectiveLevel() == logging.DEBUG



from __future__ import annotations

import logging
import warnings
from json import JSONDecodeError
from pathlib import Path
from typing import Iterable, Literal

import numpy as np
from natsort import natsorted
from ndtiff import Dataset
from numpy.typing import ArrayLike
from xarray import DataArray
from xarray import Dataset as XDataset

from iohub.mm_fov import MicroManagerFOV, MicroManagerFOVMapping

__all__ = ["NDTiffDataset", "NDTiffFOV"]
_logger = logging.getLogger(__name__)


class NDTiffFOV(MicroManagerFOV):
    def __init__(self, parent: NDTiffDataset, key: int) -> None:
        super().__init__(parent, key)
        self._xdata = parent.xdata[key]

    @property
    def axes_names(self) -> list[str]:
        return list(self._xdata.dims)

    @property
    def shape(self) -> tuple[int, int, int, int, int]:
        return self._xdata.shape

    @property
    def dtype(self) -> np.dtype:
        return self._xdata.dtype

    def __getitem__(
        self, key: int | slice | tuple[int | slice, ...]
    ) -> ArrayLike:
        return self._xdata[key]

    @property
    def xdata(self) -> DataArray:
        return self._xdata

    def frame_metadata(self, t: int, c: int, z: int) -> dict | None:
        return self.parent.get_image_metadata(self._position, t, c, z)


class NDTiffDataset(MicroManagerFOVMapping):
    """Reader for ND-TIFF datasets acquired with Micro/Pycro-Manager,
    effectively a wrapper of the `ndtiff.Dataset` class.
    """

    _ndtiff_axes: tuple[str] = ("position", "time", "channel", "z", "y", "x")

    def __init__(self, data_path: Path | str):
        super().__init__()
        data_path = Path(data_path)
        if not data_path.is_dir():
            raise FileNotFoundError(
                f"{data_path} is not a valid NDTiff dataset."
            )
        self.dataset = Dataset(str(data_path))
        self._root = data_path
        self.dirname = data_path.name
        self._axes = self.dataset.axes
        if any([a for a in self._axes.keys() if a not in self._ndtiff_axes]):
            raise NotImplementedError(
                f"Custom axis names {self._axes.keys()} are not supported. "
                f"Supported axes are: {self._ndtiff_axes}"
            )
        self._str_posistion_axis = self._check_str_axis("position")
        self._str_channel_axis = self._check_str_axis("channel")
        self.frames = (
            len(self._axes["time"]) if "time" in self._axes.keys() else 1
        )
        self.channels = (
            len(self._axes["channel"]) if "channel" in self._axes.keys() else 1
        )
        self.slices = len(self._axes["z"]) if "z" in self._axes.keys() else 1
        self.height = self.dataset.image_height
        self.width = self.dataset.image_width
        self.dtype = self.dataset.dtype

        self._all_position_keys = self._parse_all_position_keys()
        self._ndtiff_channel_names = list(self._axes.get("channel", [None]))
        self._mm_meta = self._get_summary_metadata()
        self.channel_names = self._ndtiff_channel_names
        if self.channel_names[0] is None or self.channel_names[0] == 0:
            self.channel_names = [f"Channel{i}" for i in range(self.channels)]
            _logger.warning(
                "No channel names found in metadata. Using defaults: "
                f"{self.channel_names}"
            )
        self.stage_positions = self._mm_meta["Summary"]["StagePositions"]
        z_step_size = float(self._mm_meta["Summary"]["z-step_um"] or 1.0)
        xy_pixel_size = float(
            self._mm_meta["Summary"].get("PixelSize_um") or 1.0
        )
        self._zyx_scale = (z_step_size, xy_pixel_size, xy_pixel_size)
        self._gather_xdata()

    def __enter__(self) -> NDTiffDataset:
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    def __iter__(self) -> Iterable[tuple[str, NDTiffFOV]]:
        for key in self.xdata.keys():
            key = str(key)
            yield key, NDTiffFOV(self, key)

    def __contains__(self, key: str | int) -> bool:
        return str(key) in self._xdata

    def __len__(self) -> int:
        return len(self._all_position_keys)

    def __getitem__(self, key: int | str) -> NDTiffFOV:
        return NDTiffFOV(self, str(key))

    def close(self) -> None:
        self.dataset.close()

    @property
    def zyx_scale(self) -> tuple[float, float, float]:
        return self._zyx_scale

    @property
    def t_scale(self) -> float:
        _logger.warning(
            "NDTiff does not store the planned time interval. "
            "Returning 1.0 as a placeholder."
        )
        return 1.0

    def _get_summary_metadata(self):
        pm_metadata = self.dataset.summary_metadata
        pm_metadata["MicroManagerVersion"] = "pycromanager"
        pm_metadata["Positions"] = len(self)

        c_idx = self._ndtiff_channel_names[0]
        # Get the first position index that has metadata
        for p_idx in self._all_position_keys:
            img_metadata = self.get_image_metadata(p_idx, 0, c_idx, 0)
            if img_metadata is not None:
                break

        try:
            z0 = self.get_image_metadata(p_idx, 0, c_idx, 0)[
                "ZPosition_um_Intended"
            ]
            z1 = self.get_image_metadata(p_idx, 0, c_idx, 1)[
                "ZPosition_um_Intended"
            ]
            pm_metadata["z-step_um"] = np.around(
                abs(z1 - z0), decimals=3
            ).astype(float)
        # Will raise KeyError if dataset does not have z slices
        # Will raise ValueError if dataset has only one z slice
        except (KeyError, ValueError):
            pm_metadata["z-step_um"] = None

        pm_metadata["StagePositions"] = []
        if "position" in self._axes:
            for position in self._axes["position"]:
                position_metadata = {}
                img_metadata = self.get_image_metadata(position, 0, c_idx, 0)

                if img_metadata is not None and all(
                    key in img_metadata.keys()
                    for key in [
                        "XPosition_um_Intended",
                        "YPosition_um_Intended",
                    ]
                ):
                    xy_stage = img_metadata["Core-XYStage"]
                    position_metadata["DefaultXYStage"] = xy_stage
                    position_metadata[xy_stage] = (
                        img_metadata["XPosition_um_Intended"],
                        img_metadata["YPosition_um_Intended"],
                    )

                # Position label may also be obtained from "PositionName"
                # metadata key with AcqEngJ >= 0.29.0
                if isinstance(position, str):
                    position_metadata["Label"] = position

                pm_metadata["StagePositions"].append(position_metadata)

        return {"Summary": pm_metadata}

    def _check_str_axis(self, axis: Literal["position", "channel"]) -> bool:
        if axis in self._axes:
            coord_sample = next(iter(self._axes[axis]))
            return isinstance(coord_sample, str)
        else:
            return False

    @property
    def str_position_axis(self) -> bool:
        """Position axis is string-valued"""
        return self._str_posistion_axis

    @property
    def str_channel_axis(self) -> bool:
        """Channel axis is string-valued"""
        return self._str_channel_axis

    def _check_coordinates(self, p: int | str, t: int, c: int | str, z: int):
        """
        Check that the (p, t, c, z) coordinates are part of the ndtiff dataset.
        Replace coordinates with None or string values in specific cases - see
        below
        """
        coords = [p, t, c, z]
        axes = self._ndtiff_axes[:4]

        for i, axis in enumerate(axes):
            coord = coords[i]

            # Check if the axis is part of the dataset axes
            if axis in self._axes.keys():
                # Check if coordinate is part of the dataset axis
                if coord in self._axes[axis]:
                    # all good
                    pass

                # The requested coordinate is not part of the axis
                else:
                    # If coord=0 is requested and the coordinate axis exists,
                    # but is string valued (e.g. {'Pos0', 'Pos1'}), a warning
                    # will be raised and the coordinate will be replaced by a
                    # random sample.

                    # Coordinates are in sets, here we get one sample from the
                    # set without removing it:
                    # https://stackoverflow.com/questions/59825
                    coord_sample = next(iter(self._axes[axis]))
                    if coord == 0 and isinstance(coord_sample, str):
                        coords[i] = coord_sample
                        warnings.warn(
                            f"Indices of {axis} are string-valued. "
                            f"Returning data at {axis} = {coord}"
                        )
                    else:
                        # If the coordinate is not part of the axis and
                        # nonzero, a ValueError will be raised
                        raise ValueError(
                            f"Image coordinate {axis} = {coord} is not "
                            "part of this dataset."
                        )

            # The axis is not part of the dataset axes
            else:
                # Nothing to do if coord == None
                if coord is not None:
                    # If coord = 0 is requested, the coordinate will be
                    # replaced with None
                    if coord == 0:
                        coords[i] = None
                    # If coord != 0 is requested and the axis is not part of
                    # the dataset, ValueError will be raised
                    else:
                        raise ValueError(
                            f"Axis {axis} is not part of this dataset"
                        )

        return tuple(coords)

    def _parse_all_position_keys(self) -> list[int | str | None]:
        return natsorted(list(self._axes.get("position", [None])))

    def _check_position_key(self, key: int | str) -> bool:
        if "position" in self._axes.keys():
            if key not in self._axes["position"]:
                raise ValueError(
                    f"Position index {key} is not part of this dataset. "
                    f'Valid positions are: {self._axes["position"]}'
                )
        else:
            if key not in (0, None):
                warnings.warn(
                    f"Position index {key} is not part of this dataset. "
                    "Returning data at the default position."
                )
                key = None
        return key

    def _gather_xdata(self) -> None:
        shape = (
            len(self),
            self.frames,
            self.channels,
            self.slices,
            self.height,
            self.width,
        )
        if self._all_position_keys == [None]:
            pkeys = ["0"]
        else:
            pkeys = [str(k) for k in self._all_position_keys]
        # add singleton axes so output is always (p, t, c, z, y, x)
        da = DataArray(
            self.dataset.as_array().reshape(shape),
            dims=self._ndtiff_axes,
            name=self.dirname,
            coords={"position": pkeys, "channel": self.channel_names},
        )
        self._xdata = da.to_dataset(dim="position")

    @property
    def xdata(self) -> XDataset:
        return self._xdata

    def get_image_metadata(
        self, p: int | str, t: int, c: int | str, z: int
    ) -> dict | None:
        """Return image plane metadata at the requested PTCZ coordinates

        Parameters
        ----------
        p : int or str
            position index
        t : int
            time index
        c : int or str
            channel index
        z : int
            slice/z index

        Returns
        -------
        dict | None
            Image plane metadata. None if not available.
        """
        metadata = None
        if not self.str_position_axis and isinstance(p, str):
            if p.isdigit():
                p = int(p)
        try:
            p, t, c, z = self._check_coordinates(p, t, c, z)
        except ValueError as e:
            _logger.debug(
                f"Error checking coordinates at position {p}, time {t}, "
                f"channel {c}, z {z}: {e}"
            )
        if self.dataset.has_image(position=p, time=t, channel=c, z=z):
            try:
                metadata = self.dataset.read_metadata(
                    position=p, time=t, channel=c, z=z
                )
            except (JSONDecodeError, UnicodeDecodeError):
                # acquisition crashed before metadata was written
                _logger.warning(
                    f"Unable to decode metadata for position {p}, "
                    f"time {t}, channel {c}, z {z}"
                )
        return metadata



"""
Coordinate Transform
====================

This script writes two positions using the high content screening (HCS)
OME-Zarr dataset with two FOV in a single well with different
coordinate transformations (i.e translation and scaling)
"""

# %%
import os
from tempfile import TemporaryDirectory

import numpy as np

from iohub.ngff import TransformationMeta, open_ome_zarr

# %%
# Set storage path

tmp_dir = TemporaryDirectory()
store_path = os.path.join(tmp_dir.name, "transformed.zarr")
print("Zarr store path", store_path)

# %%
# Create two random sample images
tczyx_1 = np.random.randint(
    0, np.iinfo(np.uint16).max, size=(1, 3, 3, 32, 32), dtype=np.uint16
)
tczyx_2 = np.random.randint(
    0, np.iinfo(np.uint16).max, size=(1, 3, 3, 32, 32), dtype=np.uint16
)

# %%
# Coordinate Transformations (T,C,Z,Y,X)
# By default the translation is the identity matrix
coords_shift = [[1.0, 1.0, 1.0, 10.0, 10.0], [1.0, 1.0, 0.0, -10.0, -10.0]]
img_scaling = [[1.0, 1.0, 1.0, 0.5, 0.5]]

# %%
# Generate Transformation Metadata
translation = []
for shift in coords_shift:
    translation.append(
        TransformationMeta(type="translation", translation=shift)
    )
scaling = []
for scale in img_scaling:
    scaling.append(TransformationMeta(type="scale", scale=scale))

# %%
# Write 5D data to a new Zarr store
with open_ome_zarr(
    store_path,
    layout="hcs",
    mode="w-",
    channel_names=["DAPI", "GFP", "Brightfield"],
) as dataset:
    # Create and write to positions
    # This affects the tile arrangement in visualization
    position = dataset.create_position("0", "0", "0")
    position.create_image("0", tczyx_1, transform=[translation[0]])
    position = dataset.create_position("0", "0", "1")
    position.create_image("0", tczyx_2, transform=[translation[1], scaling[0]])
    # Print dataset summary
    dataset.print_tree()

# %%
# .. note:: To see the coordinate transforms,
#     open the positions individually using napari-ome-zarr.
#     This will duplicate the layers (channels).

# %%
# Clean up
tmp_dir.cleanup()



import inspect
import itertools
import multiprocessing as mp
from collections import defaultdict
from functools import partial
from pathlib import Path
from typing import Any, Callable, Literal, Sequence, Union
from warnings import warn

import click
import numpy as np
from numpy.typing import DTypeLike, NDArray

from iohub.ngff import open_ome_zarr
from iohub.ngff.nodes import TransformationMeta


def create_empty_plate(
    store_path: Path,
    position_keys: list[tuple[str, str, str]],
    channel_names: list[str],
    shape: tuple[int, ...],
    chunks: tuple[int, ...] | None = None,
    shards_ratio: tuple[int, ...] | None = None,
    version: Literal["0.4", "0.5"] = "0.4",
    scale: tuple[float, ...] = (1, 1, 1, 1, 1),
    dtype: DTypeLike = np.float32,
    max_chunk_size_bytes: float = 500e6,
) -> None:
    """
    Create a new HCS Plate in OME-Zarr format if the plate does not exist.
    If the plate exists, append positions and channels
    if they are not already in the plate.

    Parameters
    ----------
    store_path : Path
        Path to the HCS plate.
    position_keys : list[tuple[str, str, str]]
        Position keys (row, column, fov) to append if not present in the plate,
        e.g., [("A", "1", "0"), ("A", "1", "1")].
    channel_names : list[str]
        List of channel names. If the store exists,
        append if not present in metadata.
    shape : tuple[int, ...]
        TCZYX shape of the plate.
    chunks : tuple[int, ...], optional
        TCZYX chunk size of the plate. If None, the chunk size is calculated
        based on the shape to be less than max_chunk_size_bytes.
        Defaults to None.
    shards_ratio : tuple[int, ...], optional
        TCZYX shards ratio of the plate.
        If None, no sharding is applied.
        Defaults to None.
    version : Literal["0.4", "0.5"], optional
        OME-Zarr version to use for the plate.
        Defaults to "0.4".
    scale : tuple[float, ...], optional
        TCZYX scale of the plate. Defaults to (1, 1, 1, 1, 1).
    dtype : DTypeLike, optional
        Data type of the plate. Defaults to np.float32.
    max_chunk_size_bytes : float, optional
        Maximum chunk size in bytes. Defaults to 500e6 (500MB).

    Examples
    --------
    Create a new plate with positions and channels:
    >>> create_empty_plate(
    ...     store_path=Path("/path/to/store"),
    ...     position_keys=[("A", "1", "0"), ("A", "1", "1")],
    ...     channel_names=["DAPI", "FITC"],
    ...     shape=(1, 1, 256, 256, 256)
    ... )

    Create a plate with custom chunk size and scale:
    >>> create_empty_plate(
    ...     store_path=Path("/path/to/store"),
    ...     position_keys=[("A", "1", "0")],
    ...     channel_names=["DAPI"],
    ...     shape=(1, 1, 256, 256, 256),
    ...     chunks=(1, 1, 128, 128, 128),
    ...     scale=(1, 1, 0.5, 0.5, 0.5)
    ... )

    Create a plate with sharding:
    >>> create_empty_plate(
    ...     store_path=Path("/path/to/store"),
    ...     position_keys=[("A", "1", "0")],
    ...     channel_names=["DAPI"],
    ...     shape=(1, 1, 64, 2048, 2048),
    ...     chunks=(1, 1, 8, 128, 128),
    ...     scale=(1, 1, 0.5, 0.5, 0.5),
    ...     shards_ratio=(10, 1, 8, 16, 16),
    ...     version="0.5"
    ... )

    Notes
    -----
    - If `chunks` is not provided, the function calculates an appropriate
    chunk size to keep the chunks under the specified `max_chunk_size_bytes`.
    - The function ensures that positions and channels are appended to an
    existing plate if they are not already present.
    """

    # Limiting the chunking to 500MB
    if chunks is None:
        chunk_zyx_shape = _calculate_zyx_chunk_size(
            shape,
            np.dtype(dtype).itemsize,
            max_chunk_size_bytes,
        )

        chunks = 2 * (1,) + chunk_zyx_shape

    # Create plate
    output_plate = open_ome_zarr(
        str(store_path),
        layout="hcs",
        mode="a",
        channel_names=channel_names,
        version=version,
    )
    if output_plate.version == "0.4" and shards_ratio is not None:
        warn("Ignoring shards ratio for OME-Zarr version 0.4.")
        shards_ratio = None

    # Create positions
    for position_key in position_keys:
        position_key_string = "/".join(position_key)
        # Check if position is already in the store, if not create it
        if position_key_string not in output_plate.zgroup:
            position = output_plate.create_position(*position_key)
            _ = position.create_zeros(
                name="0",
                shape=shape,
                chunks=chunks,
                shards_ratio=shards_ratio,
                dtype=dtype,
                transform=[TransformationMeta(type="scale", scale=scale)],
            )
        else:
            position = output_plate[position_key_string]

        # Check if channel_names are already in the store, if not append them
        for channel_name in channel_names:
            metadata_channel_names = position.channel_names
            if channel_name not in metadata_channel_names:
                position.append_channel(channel_name, resize_arrays=True)


def _apply_transform_to_czyx(
    func: Callable[[NDArray, Any], NDArray],
    input_position_path: Path,
    input_channel_indices: Union[list[int], slice],
    input_time_index: int,
    **kwargs,
) -> NDArray | None:
    # Check if input_time_indices should be added to the func kwargs
    # This is needed when a different processing is needed for each time point,
    # for example during stabilization
    all_func_params = inspect.signature(func).parameters.keys()
    if "input_time_index" in all_func_params:
        kwargs["input_time_index"] = input_time_index

    # Process CZYX given with the given indices
    # if input_channel_indices is not None and len(input_channel_indices) > 0:
    click.echo(f"Processing t={input_time_index}, c={input_channel_indices}")
    input_dataset = open_ome_zarr(input_position_path, layout="fov", mode="r")
    czyx_data = input_dataset.data.oindex[
        input_time_index, input_channel_indices
    ]
    if not _check_nan_n_zeros(czyx_data):
        return func(czyx_data, **kwargs)
    else:
        return None


def _echo_finished(
    time_index: int | list[int] | slice,
    channel_index: int | list[int] | slice,
    skipped: bool,
) -> None:
    if skipped:
        click.echo(
            f"Skipping t={time_index}, c={channel_index} "
            "due to all zeros or nans"
        )
    else:
        click.echo(f"Finished writing t={time_index}, c={channel_index}")


def _save_transformed(
    transformed: NDArray | list[NDArray] | None,
    output_position_path: Path,
    output_channel_indices: list[int] | slice,
    output_time_indices: int | list[int],
) -> None:
    # NOTE: use tensorstore due to zarr-python#3221
    import tensorstore

    with open_ome_zarr(
        output_position_path, layout="fov", mode="r+"
    ) as output_dataset:
        ts = output_dataset.data.tensorstore(
            context=tensorstore.Context(
                {"data_copy_concurrency": {"limit": 4}}
            )
        )
    ts.oindex[output_time_indices, output_channel_indices].write(
        transformed
    ).result()
    # NOTE: explicit GC due to tensorstore#223
    del ts


def apply_transform_to_czyx_and_save(
    func: Callable[[NDArray, Any], NDArray],
    input_position_path: Path,
    output_position_path: Path,
    input_channel_indices: Union[list[int], slice],
    output_channel_indices: Union[list[int], slice],
    input_time_index: int,
    output_time_index: int,
    **kwargs,
) -> None:
    """
    Note: To be deprecated, no longer used by process_single_position

    Load a CZYX array from a position store,
    apply a transformation, and save the result.

    Parameters
    ----------
    func : Callable[[NDArray, Any], NDArray]
        The function to be applied to the data.
        func must take the CZYX NDArray as the first argument and return
        a CZXY NDArray. Additional arguments are passed through **kwargs.
    input_position_path : Path
        The path to input OME-Zarr position store
        (e.g., input_store_path.zarr/A/1/0).
    output_position_path : Path
        The path to output OME-Zarr position store
        (e.g., output_store_path.zarr/A/1/0).
    input_channel_indices : Union[list[int], slice]
        The channel indices to process. Acceptable values:
        - Slices: slice(0, 2).
        - A list of integers: [0, 1, 2, 3, 4].
    output_channel_indices : Union[list[int], slice]
        The channel indices to write to. Acceptable values:
        - Slices: slice(0, 2).
        - A list of integers: [0, 1, 2, 3, 4].
    input_time_index : int
        The time index to process.
    output_time_index : int
        The time index to write to.
    kwargs : dict, optional
        Additional arguments to pass to the function.

    Examples
    --------
    Using slices for input_channel_indices:
    >>> apply_transform_to_zyx_and_save(
    ...     func=some_function,
    ...     input_position_path=Path("/path/to/input.zarr/A/1/0"),
    ...     output_position_path=Path("/path/to/output.zarr/A/1/0"),
    ...     input_channel_indices=slice(0, 2),
    ...     output_channel_indices=[0],
    ...     input_time_index=0,
    ...     output_time_index=0,
    ... )

    Using list for input_channel_indices:
    >>> apply_transform_to_zyx_and_save(
    ...     func=some_function,
    ...     input_position_path=Path("/path/to/input.zarr/A/1/0"),
    ...     output_store_path=Path("/path/to/output.zarr/A/1/0"),
    ...     input_channel_indices=[0, 1, 2, 3, 4],
    ...     output_channel_indices=[0, 1, 2],
    ...     input_time_index=0,
    ...     output_time_index=0,
    ... )

    """
    transformed = _apply_transform_to_czyx(
        func,
        input_position_path=input_position_path,
        input_channel_indices=input_channel_indices,
        input_time_index=input_time_index,
        **kwargs,
    )
    if transformed is not None:
        _save_transformed(
            transformed,
            output_time_indices=output_time_index,
            output_channel_indices=output_channel_indices,
            output_position_path=output_position_path,
        )
        _echo_finished(
            time_index=input_time_index,
            channel_index=input_channel_indices,
            skipped=False,
        )
    else:
        _echo_finished(
            time_index=input_time_index,
            channel_index=input_channel_indices,
            skipped=True,
        )


def _indices_to_shard_aligned_batches(
    indices: Sequence[int], shard_size: int
) -> list[list[int]]:
    """Split indices into batches that are in the same shards.

    Parameters
    ----------
    indices : Sequence[int]
        Non-negative indices to split.
    shard_size : int
        The size of each shard.

    Returns
    -------
    list[list[int]]
        List of sorted batches,
        where each batch is a list of indices in the same shard.
    """
    indices = sorted(indices)
    batches = defaultdict(list)
    for index in indices:
        if index < 0:
            raise ValueError(f"Negative indices are not supported: {indices}")
        batches[index // shard_size].append(index)
    return list(batches.values())


def _match_indices_to_batches(
    flat_indices: Sequence[int],
    original_reference: Sequence[int],
    batched_reference: list[list[int]],
) -> list[list[int]]:
    """Match flat indices to batches based on a reference pair.

    Parameters
    ----------
    flat_indices : Sequence[int]
        Flat indices to match.
    original_reference : Sequence[int]
        Original reference indices.
    batched_reference : list[list[int]]
        Batched version of reference.

    Returns
    -------
    list[list[int]]
        List of matched batches, where each batch corresponds to the
        original reference indices.
    """
    matched_batches = []
    for batch in batched_reference:
        matched_batch = []
        for index in batch:
            matched_batch.append(flat_indices[original_reference.index(index)])
        matched_batches.append(matched_batch)
    return matched_batches


def _slice_to_list(indices: list[int] | slice) -> list[int]:
    if isinstance(indices, slice):
        return list(range(indices.start, indices.stop, indices.step))
    return indices


def apply_transform_to_tczyx_and_save(
    func: Callable[[NDArray, Any], NDArray],
    input_position_path: Path,
    output_position_path: Path,
    input_channel_indices: list[int] | slice,
    output_channel_indices: list[int] | slice,
    input_time_indices: list[int] | slice,
    output_time_indices: list[int] | slice,
    **kwargs,
) -> None:
    """
    Load a TCZYX array from a position store,
    apply a transformation, and save the result.

    Parameters
    ----------
    func : Callable[[NDArray, Any], NDArray]
        The function to be applied to the data.
        func must take the TCZYX NDArray as the first argument and return
        a TCZXY NDArray. Additional arguments are passed through **kwargs.
    input_position_path : Path
        The path to input OME-Zarr position store
        (e.g., "input_store_path.zarr/A/1/0").
    output_position_path : Path
        The path to output OME-Zarr position store
        (e.g., "output_store_path.zarr/A/1/0").
    input_channel_indices : list[int] | slice
        The channel indices to process. Acceptable values:
        - Slices: slice(0, 2).
        - A list of integers: [0, 1, 2, 3, 4].
    output_channel_indices : list[int] | slice
        The channel indices to write to,
        similar to input_channel_indices.
    input_time_indices : list[int] | slice
        The time indices to process,
        similar to input_channel_indices.
    output_time_indices : list[int] | slice
        The time indices to write to,
        similar to input_channel_indices.
    kwargs : dict, optional
        Additional arguments to pass to the function.
    """
    input_time_indices = _slice_to_list(input_time_indices)
    results = {}
    for i, input_time_index in enumerate(input_time_indices):
        result = _apply_transform_to_czyx(
            func,
            input_position_path=input_position_path,
            input_channel_indices=input_channel_indices,
            input_time_index=input_time_index,
            **kwargs,
        )
        if result is not None:
            results[i] = result
        else:
            _echo_finished(
                time_index=input_time_index,
                channel_index=input_channel_indices,
                skipped=True,
            )
    if results:
        output_time_indices = _slice_to_list(output_time_indices)
        output_time_indices = [output_time_indices[i] for i in results.keys()]
        _save_transformed(
            transformed=list(results.values()),
            output_position_path=output_position_path,
            output_channel_indices=output_channel_indices,
            output_time_indices=output_time_indices,
        )
        _echo_finished(
            input_time_indices, input_channel_indices, skipped=False
        )
    del results


def process_single_position(
    func: Callable[[NDArray, Any], NDArray],
    input_position_path: Path,
    output_position_path: Path,
    input_channel_indices: list[slice] | list[list[int]] | None = None,
    output_channel_indices: list[slice] | list[list[int]] | None = None,
    input_time_indices: list[int] | None = None,
    output_time_indices: list[int] | None = None,
    num_processes: int = 1,
    **kwargs,
) -> None:
    """
    Apply function to data in an `iohub` position store,
    parallelizing over time and channel indices,
    and save result in an output position store.

    Parameters
    ----------
    func : Callable[[NDArray, Any], NDArray]
        The function to be applied to the data.
        func must take the CZYX NDArray as the first argument and return
        a CZXY NDArray. Additional arguments are passed through **kwargs.
    input_position_path : Path
        The path to the input OME-Zarr position store
        (e.g., input_store_path.zarr/A/1/0).
    output_position_path : Path
        The path to the output OME-Zarr position store
        (e.g., output_store_path.zarr/A/1/0).
    input_time_indices : list[int], optional
        If not provided, all timepoints will be processed.
    output_time_indices : list[int], optional
        The time indices to write to. Must match length of input_time_indices.
        Typically used for stabilization, which needs per timepoint processing.
        If not provided, input_time_indices will be used.
    input_channel_indices : Union[list[slice], list[list[int]]], optional
        The channel indices to process. Acceptable values:
        - A list of slices: [slice(0, 2), slice(2, 4), ...].
        - A list of lists of integers: [[0, 1, 2, 3, 4]].
        If empty, process all channels.
        Must match output_channel_indices if not empty.
        Defaults to None.
    output_channel_indices : Union[list[slice], list[list[int]]], optional
        The channel indices to write to. Acceptable values:
        - A list of slices: [slice(0, 2), slice(2, 4), ...].
        - A list of lists of integers: [[0, 1, 2, 3, 4]].
        If empty, write to all channels.
        Must match input_channel_indices if not empty.
        Defaults to None.
    num_processes : int, optional
        Number of simultaneous processes per position. Defaults to 1.
    kwargs : dict, optional
        Additional arguments to pass to the function.
        A dictionary with key "extra_metadata"
        can be passed to be stored at a FOV level,
        e.g.,
        kwargs={"extra_metadata": {"Temperature": 37.5, "CO2_level": 0.5}}.

    Examples
    --------
    Using slices for input_channel_indices:
    process_single_position(
        func=some_function,
        input_position_path=Path("/path/to/input"),
        output_store_path=Path("/path/to/output"),
        input_time_indices=[0],
        input_channel_indices=[slice(0, 2), slice(2, 4)],
        output_channel_indices=[[0], [1]],
    )

    Using list of lists for input_channel_indices:
    process_single_position(
        func=some_function,
        input_position_path=Path("/path/to/input"),
        output_store_path=Path("/path/to/output"),
        input_time_indices=[1, 2],
        input_channel_indices=[[0, 1], [2, 3]],
        output_channel_indices=[[0], [1]],
    )

    Notes
    -----
    - Multiprocessing over T and C:
    input_channel_indices and output_channel_indices should be empty.
    - Multiprocessing over T only:
    input_channel_indices and output_channel_indices should be provided.
    """
    # Function to be applied
    click.echo(f"Function to be applied: \t{func}")
    click.echo(f"Input data path:\t{input_position_path}")
    click.echo(f"Output data path:\t{output_position_path}")

    # Get the reader
    with open_ome_zarr(
        input_position_path, layout="fov", mode="r"
    ) as input_dataset:
        input_data_shape = input_dataset.data.shape
    with open_ome_zarr(
        output_position_path, layout="fov", mode="r"
    ) as output_dataset:
        output_shards = output_dataset.data.shards

    # Process time indices
    if input_time_indices is None:
        input_time_indices = list(range(input_data_shape[0]))
    assert (
        type(input_time_indices) is list
    ), "input_time_indices must be a list"
    if output_time_indices is None:
        output_time_indices = input_time_indices
    if output_shards is not None:
        batched_output_time_indices = _indices_to_shard_aligned_batches(
            output_time_indices, output_shards[0]
        )
        batched_input_time_indices = _match_indices_to_batches(
            flat_indices=input_time_indices,
            original_reference=output_time_indices,
            batched_reference=batched_output_time_indices,
        )
    else:
        batched_input_time_indices = [[i] for i in input_time_indices]
        batched_output_time_indices = [[i] for i in output_time_indices]

    # Process channel indices
    if input_channel_indices is None:
        input_channel_indices = [[c] for c in range(input_data_shape[1])]
        output_channel_indices = input_channel_indices
    assert (
        type(input_channel_indices) is list
    ), "input_channel_indices must be a list"
    if output_channel_indices is None:
        output_channel_indices = input_channel_indices
    if output_shards is not None and output_shards[1] != 1:
        raise ValueError(
            "Sharding along the channel dimension is not supported."
        )

    # Check for invalid times
    time_ubound = input_data_shape[0] - 1
    if np.max(input_time_indices) > time_ubound:
        raise ValueError(
            f"""input_time_indices = {input_time_indices} includes
            a time index beyond the maximum index of
            the dataset = {time_ubound}"""
        )

    # Write extra metadata to the output store
    extra_metadata = kwargs.pop("extra_metadata", None)
    with open_ome_zarr(
        output_position_path, layout="fov", mode="r+"
    ) as output_dataset:
        output_dataset.zattrs["extra_metadata"] = extra_metadata

    # Loop through (T, C), applying transform and writing as we go
    iterable = itertools.product(
        zip(input_channel_indices, output_channel_indices),
        zip(batched_input_time_indices, batched_output_time_indices),
    )
    flat_iterable = tuple((*c, *t) for c, t in iterable)

    partial_apply_transform_to_czyx_and_save = partial(
        apply_transform_to_tczyx_and_save,
        func,
        input_position_path,
        output_position_path,
        **kwargs,
    )
    num_processes = min(num_processes, len(flat_iterable), mp.cpu_count())
    click.echo(f"\nStarting multiprocess pool with {num_processes} processes")
    # NOTE: use spawn to work around tensorstore#61
    context = mp.get_context("spawn")
    with context.Pool(num_processes) as p:
        p.starmap(
            partial_apply_transform_to_czyx_and_save,
            flat_iterable,
        )
    click.echo("Shut down multiprocess pool")


def _check_nan_n_zeros(input_array) -> bool:
    """
    Checks if any of the channels are all zeros or nans and returns true
    """
    if len(input_array.shape) == 3:
        # Check if all the values are zeros or nans
        if np.all(input_array == 0) or np.all(np.isnan(input_array)):
            # Return true
            return True
    elif len(input_array.shape) == 4:
        # Get the number of channels
        num_channels = input_array.shape[0]
        # Loop through the channels
        for c in range(num_channels):
            # Get the channel
            zyx_array = input_array[c, :, :, :]

            # Check if all the values are zeros or nans
            if np.all(zyx_array == 0) or np.all(np.isnan(zyx_array)):
                # Return true
                return True
    else:
        raise ValueError("Input array must be 3D or 4D")

    # Return false
    return False


def _calculate_zyx_chunk_size(
    shape, bytes_per_pixel, max_chunk_size_bytes
) -> tuple[int, int, int]:
    """
    Calculate the chunk size for ZYX dimensions based on the shape,
    bytes per pixel of data, and desired max chunk size.
    """

    chunk_zyx_shape = list(shape[-3:])

    # while XY image is larger than MAX_CHUNK_SIZE
    while (
        chunk_zyx_shape[-3] > 1
        and np.prod(chunk_zyx_shape) * bytes_per_pixel > max_chunk_size_bytes
    ):
        chunk_zyx_shape[-3] = np.ceil(chunk_zyx_shape[-3] / 2)
    chunk_zyx_shape = tuple(map(int, chunk_zyx_shape))
    return chunk_zyx_shape



from __future__ import annotations

import logging
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import natsort
import tifffile as tiff
import zarr

from iohub._deprecated.reader_base import ReaderBase
from iohub._deprecated.singlepagetiff import MicromanagerSequenceReader
from iohub._deprecated.zarrfile import ZarrReader
from iohub.fov import BaseFOVMapping
from iohub.mmstack import MMStack
from iohub.ndtiff import NDTiffDataset
from iohub.ngff.nodes import NGFFNode, Plate, Position, open_ome_zarr

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath

_logger = logging.getLogger(__name__)


def _find_ngff_version_in_zarr_group(group: zarr.Group) -> str | None:
    for key in ["plate", "well", "ome"]:
        if key in group.attrs:
            if v := group.attrs[key].get("version"):
                return v
    if "multiscales" in group.attrs:
        for ms in group.attrs["multiscales"]:
            if v := ms.get("version"):
                return v
    return None


def _check_zarr_data_type(src: Path):
    try:
        root = zarr.open(src, mode="r")
        if version := _find_ngff_version_in_zarr_group(root):
            return version
        else:
            for _, row in root.groups():
                for _, well in row.groups():
                    if version := _find_ngff_version_in_zarr_group(well):
                        return version
    except Exception:
        return False
    return "unknown"


def _check_single_page_tiff(src: Path):
    if src.is_file():
        src = src.parent
    files = src.glob("*.tif")
    try:
        next(files)
    except StopIteration:
        sub_dirs = _get_sub_dirs(src)
        if sub_dirs:
            files = (src / sub_dirs[0]).glob("*.tif")
            try:
                with tiff.TiffFile(next(files)) as tf:
                    if (
                        len(tf.pages) == 1
                    ):  # and tf.pages[0].is_multipage is False:
                        return True
            except StopIteration:
                pass
    return False


def _check_multipage_tiff(src: Path):
    if src.is_file():
        src = src.parent
    try:
        file = next(src.glob("*.tif"))
    except StopIteration:
        return False
    with tiff.TiffFile(file) as tf:
        if len(tf.pages) > 1:
            return True
        elif tf.is_multipage is False and tf.is_ome is True:
            return True
    return False


def _check_ndtiff(src: Path):
    # select parent directory if a .tif file is selected
    if src.is_file() and "tif" in src.suffixes:
        if _check_ndtiff(src.parent) or _check_ndtiff(src.parent.parent):
            return True
    # ndtiff v2
    if Path(src, "Full resolution", "NDTiff.index").exists():
        return True
    # ndtiff v3
    elif Path(src, "NDTiff.index").exists():
        return True
    return False


def _get_sub_dirs(directory: Path) -> list[str]:
    """ """
    sub_dir_name = [
        subdir.name for subdir in directory.iterdir() if subdir.is_dir()
    ]
    #    assert subDirName, 'No sub directories found'
    return natsort.natsorted(sub_dir_name)


def _infer_format(path: Path):
    extra_info = None
    if ngff_version := _check_zarr_data_type(path):
        data_type = "omezarr"
        extra_info = ngff_version
    elif _check_ndtiff(path):
        data_type = "ndtiff"
    elif _check_multipage_tiff(path):
        data_type = "ometiff"
    elif _check_single_page_tiff(path):
        data_type = "singlepagetiff"
    else:
        raise RuntimeError(
            "Failed to infer data type: "
            f"No compatible data found under {path}."
        )
    return (data_type, extra_info)


def read_images(
    path: StrOrBytesPath,
    data_type: Literal[
        "singlepagetiff", "ometiff", "ndtiff", "omezarr"
    ] = None,
) -> ReaderBase | BaseFOVMapping:
    """Read image arrays and metadata from a Micro-Manager dataset.
    Supported formats are Micro-Manager-acquired TIFF datasets
    (single-page TIFF, multi-page OME-TIFF, NDTIFF),
    and converted OME-Zarr (v0.1/v0.4 HCS layout assuming a linear structure).

    Parameters
    ----------
    path : StrOrBytesPath
        File path, directory path to ome-tiff series, or Zarr root path
    data_type :
    Literal["singlepagetiff", "ometiff", "ndtiff", "omezarr"], optional
        Dataset format, by default None

    Returns
    -------
    ReaderBase | BaseFOVMapping
        Image collection object for the dataset
    """
    path = Path(path).resolve()
    # try to guess data type
    extra_info = None
    if not data_type:
        data_type, extra_info = _infer_format(path)
    _logger.debug(f"Detected data type: {data_type} {extra_info}")
    # identify data structure type
    if data_type == "ometiff":
        return MMStack(path)
    elif data_type == "singlepagetiff":
        return MicromanagerSequenceReader(path)
    elif data_type == "omezarr":
        if extra_info is None:
            _, extra_info = _infer_format(path)
        if extra_info == "0.4":
            # `warnings` instead of `logging` since this can be avoided
            warnings.warn(
                UserWarning(
                    "For NGFF v0.4 datasets, `iohub.open_ome_zarr()` "
                    "is preferred over `iohub.read_images()`. "
                    "Note that `open_ome_zarr()` will return "
                    "an NGFFNode object instead of a ReaderBase instance."
                )
            )
        elif extra_info != "0.1":
            raise ValueError(f"NGFF version {extra_info} is not supported.")
        return ZarrReader(path, version=extra_info)
    elif data_type == "ndtiff":
        return NDTiffDataset(path)
    else:
        raise ValueError(f"Reader of type {data_type} is not implemented")


def print_info(path: StrOrBytesPath, verbose=False):
    """Print summary information for a dataset.

    Parameters
    ----------
    path : StrOrBytesPath
        Path to the dataset
    verbose : bool, optional
        Show usage guide to open dataset in Python
        and full tree for HCS Plates in OME-Zarr,
        by default False
    """
    path = Path(path).resolve()
    try:
        fmt, extra_info = _infer_format(path)
        if fmt == "omezarr" and extra_info in ("0.4", "0.5"):
            reader = open_ome_zarr(path, mode="r", version=extra_info)
        else:
            reader = read_images(path, data_type=fmt)
    except (ValueError, RuntimeError):
        print("Error: No compatible dataset is found.")
        return
    fmt_msg = f"Format:\t\t\t {fmt}"
    if extra_info:
        if extra_info.startswith("0."):
            fmt_msg += " v" + extra_info
    sum_msg = "\n=== Summary ==="
    ch_msg = f"Channel names:\t\t {reader.channel_names}"
    code_msg = "\nThis datset can be opened with iohub in Python code:\n"
    msgs = []
    if isinstance(reader, BaseFOVMapping):
        _, first_fov = next(iter(reader))
        shape_msg = ", ".join(
            [
                f"{a}={s}"
                for s, a in zip(first_fov.shape, ("T", "C", "Z", "Y", "X"))
            ]
        )
        msgs.extend(
            [
                sum_msg,
                fmt_msg,
                f"FOVs:\t\t\t {len(reader)}",
                f"FOV shape:\t\t {shape_msg}",
                ch_msg,
                f"(Z, Y, X) scale (um):\t {first_fov.zyx_scale}",
            ]
        )
        if verbose:
            msgs.extend(
                [
                    code_msg,
                    ">>> from iohub import read_images",
                    f">>> reader = read_images('{path}')",
                ]
            )
        print(str.join("\n", msgs))
    elif isinstance(reader, NGFFNode):
        msgs.extend(
            [
                sum_msg,
                fmt_msg,
                "".join(
                    ["Axes:\t\t\t "]
                    + [f"{a.name} ({a.type}); " for a in reader.axes]
                ),
                ch_msg,
            ]
        )
        if isinstance(reader, Plate):
            meta = reader.metadata
            msgs.extend(
                [
                    f"Row names:\t\t {[r.name for r in meta.rows]}",
                    f"Column names:\t\t {[c.name for c in meta.columns]}",
                    f"Wells:\t\t\t {len(meta.wells)}",
                ]
            )
            if verbose:
                print("Zarr hierarchy:")
                reader.print_tree()
                positions = list(reader.positions())
                total_bytes_uncompressed = sum(
                    p["0"].nbytes for _, p in positions
                )
                msgs.append(f"Positions:\t\t {len(positions)}")
                msgs.append(f"Chunk size:\t\t {positions[0][1][0].chunks}")
                msgs.append(
                    f"No. bytes decompressed:\t\t {total_bytes_uncompressed} "
                    f"[{sizeof_fmt(total_bytes_uncompressed)}]"
                )
        else:
            total_bytes_uncompressed = reader["0"].nbytes
            msgs.append(f"(Z, Y, X) scale (um):\t {tuple(reader.scale[2:])}")
            msgs.append(f"Chunk size:\t\t {reader['0'].chunks}")
            msgs.append(
                f"No. bytes decompressed:\t\t {total_bytes_uncompressed} "
                f"[{sizeof_fmt(total_bytes_uncompressed)}]"
            )
        if verbose:
            msgs.extend(
                [
                    code_msg,
                    ">>> from iohub import open_ome_zarr",
                    f">>> dataset = open_ome_zarr('{path}', mode='r')",
                ]
            )
        if isinstance(reader, Position):
            print("Zarr hierarchy:")
            reader.print_tree()
        print("\n".join(msgs))
        reader.close()


def sizeof_fmt(num: int) -> str:
    """
    Human readable file size
    Adapted form:
    https://web.archive.org/web/20111010015624/
    http://blogmag.net/blog/read/38/Print_human_readable_file_size
    """
    if num < 1024:
        return f"{num} B"
    for x in ["KiB", "MiB", "GiB", "TiB"]:
        num /= 1024
        if num < 1024:
            return f"{num:.1f} {x}"



# libraries for singlepage tiff sequence reading
import glob
import json
import logging
import os
from pathlib import Path

import natsort
import numpy as np
import tifffile as tiff
import zarr

from iohub._deprecated.reader_base import ReaderBase
from iohub.mmstack import _tiff_to_fsspec_store


class MicromanagerSequenceReader(ReaderBase):
    def __init__(self, folder, extract_data=False, strict=False):
        super().__init__()

        """
        reads single-page tiff files generated by micro-manager
        into zarr or numpy arrays
        Strategy:
            1. Gather summary metadata from any metadata.txt
            2. Build a map between image coordinates and image file names
            3. Upon call to extract data, assign each scene to self.positions
                (similar to OmeTiffReader)

        Parameters
        ----------
        folder:         (str)
            folder containing position subdirectories,
            which contain singlepage tiff sequences
        extract_data    (bool)
            True if zarr arrays should be extracted immediately
        strict:         (bool)
            True if failures in getting images should raise exceptions
        """

        if not os.path.isdir(folder):
            raise NotImplementedError(
                "supplied path for singlepage tiff sequence reader "
                "is not a folder"
            )
        self._strict = strict

        self.log = logging.getLogger(__name__)
        self.positions = {}
        self.num_positions = None
        self.z_step_size = None
        self.height = 0
        self.width = 0
        self.frames = 0
        self.slices = 0
        self.channels = 0
        self.channel_names = []

        self.coord_to_filename = {}

        # identify type of subdirectory
        sub_dirs = self._get_sub_dirs(folder)
        if sub_dirs:
            pos_path = os.path.join(folder, sub_dirs[0])
        else:
            raise AttributeError(
                "supplied folder does not contain position "
                "or default subdirectories"
            )

        self._set_mm_meta(pos_path)

        # create coordinate to filename maps
        self.coord_to_filename = self.read_tiff_series(folder)

        # update coordinates if the acquisition finished early
        self._dims_from_coordinates()

        # todo: consider iterating over all positions.
        # Doable once we add a metadata search for stage positions
        if extract_data:
            self._create_stores(0)

    def _set_mm_meta(self, one_pos):
        """
        assign image metadata from summary metadata

        Parameters
        ----------
        one_pos:        (str) path to one position subfolder

        Returns
        -------

        """
        # pull one metadata sample and extract experiment dimensions
        metadata_path = os.path.join(one_pos, "metadata.txt")
        with open(metadata_path, "r") as f:
            self._mm_meta = json.load(f)

        mm_version = self._mm_meta["Summary"]["MicroManagerVersion"]
        if mm_version == "1.4.22":
            self._mm1_meta_parser()
        elif "beta" in mm_version:
            self._mm2beta_meta_parser()
        elif "gamma" in mm_version:
            self._mm2gamma_meta_parser()
        else:
            try:
                self._mm2gamma_meta_parser()
            except Exception:
                raise RuntimeError(
                    "Current MicroManager reader only supports version "
                    f"1.4.22 and 2.0 but {mm_version} was detected. "
                    "Metadata parsing failed."
                )

    def get_zarr(self, position):
        """
        return a zarr array for a given position

        Parameters
        ----------
        position:       (int) position (aka ome-tiff scene)

        Returns
        -------
        position:       (zarr.array)

        """
        if position not in self.positions.keys():
            self._create_stores(position)
        return self.positions[position]

    def get_image(self, p: int, t: int, c: int, z: int):
        zarray = self.get_zarr(p)
        dim_slices = [slice(None)] * 3
        for i, (dim, idx) in enumerate(
            zip((self.frames, self.channels, self.slices), (t, c, z))
        ):
            if dim > 0:
                dim_slices[i] = slice(idx, idx + 1)

        image = zarray[tuple(dim_slices)]
        return np.squeeze(image)

    def get_array(self, position):
        """
        return a numpy array for a given position

        Parameters
        ----------
        position:   (int) position (aka ome-tiff scene)

        Returns
        -------
        position:   (np.ndarray)

        """
        if position not in self.positions.keys():
            self.log.info(
                f"position {position} not yet extracted, extracting ..."
            )
            self._create_stores(position)
        return np.array(self.positions[position])

    def get_num_positions(self):
        """
        get total number of scenes referenced in ome-tiff metadata

        Returns
        -------
        number of positions     (int)

        """
        # self.log.warning("num positions for singlepage tiff reader is ambiguous.only loaded positions are reported") # noqa
        if self.positions:
            return self.num_positions
            # return len(self.positions)
        else:
            self.log.error("singlepage tiffs not loaded")

    def _create_stores(self, p):
        """
        extract all singlepage tiffs at each coordinate
        and place them in a zarr array
        coordinates are of shape = (pos, time, channel, z)
        arrays are of shape = (time, channel, z, height, width)

        Parameters
        ----------
        p:      (int) position (aka ome-tiff scene) to extract

        Returns
        -------

        """

        self.log.info("")
        z = zarr.zeros(
            shape=(
                self.frames,
                self.channels,
                self.slices,
                self.height,
                self.width,
            ),
            chunks=(1, 1, 1, self.height, self.width),
        )
        for c, fn in self.coord_to_filename.items():
            if c[0] == p:
                self.log.info(f"reading coord = {c} from filename = {fn}")
                with tiff.imread(fn, aszarr=True) as store:
                    try:
                        array = zarr.open(
                            _tiff_to_fsspec_store(
                                store, root_uri=Path(fn).parent.as_uri()
                            ),
                            mode="r",
                        )[:]
                        z[c[1], c[2], c[3]] = array
                    except Exception:
                        self.log.error(
                            f"error reading file {fn} for coordinate {c}"
                        )

        # check that the array was assigned
        if self._strict:
            if z == zarr.zeros(
                shape=(
                    self.frames,
                    self.channels,
                    self.slices,
                    self.height,
                    self.width,
                ),
                chunks=(1, 1, 1, self.height, self.width),
            ):
                raise IOError(f"array at position {p} can not be found")

        self.positions[p] = z

    def read_tiff_series(self, folder: str):
        """
        given a folder containing position subfolders, each of which contains
        single-page-tiff series acquired using micro-manager,
        parse the metadata to map image coordinates to filepaths/names

        Parameters
        ----------
        folder:             (str)
            project folder containing all position subfolders

        Returns
        -------
        coord_filename_map  (dict)
            keys are coordinates and values are filenames.
            Coordinates follow (p, t, c, z) indexing.
        """

        positions = [
            p
            for p in os.listdir(folder)
            if os.path.isdir(os.path.join(folder, p))
        ]
        if not positions:
            raise FileNotFoundError(
                "no position subfolder found in supplied folder"
            )

        metadatas = [
            os.path.join(folder, position, "metadata.txt")
            for position in positions
        ]
        if not metadatas:
            raise FileNotFoundError(
                "no metadata.txt file found in position directories"
            )

        coord_filename_map = {}
        for idx, metadata in enumerate(metadatas):
            with open(metadata, "r") as m:
                j = json.load(m)
                coord_filename_map.update(
                    self._extract_coord_to_filename(j, folder, positions[idx])
                )
        self.num_positions = len(positions)

        return coord_filename_map

    def _extract_coord_to_filename(self, json_, parent_folder, position):
        """
        given a micro-manager generated metadata json,
        extract image coordinates and their corresponding image filepaths
        build a mapping between the two.

        Parameters
        ----------
        json_:              (dict)
            dict generated from json.load of mm metadata
        parent_folder:      (str)
            full path to file
        position:           (str)
            mm1.4.22 metadata does not associate
            positions with images in the metadata.
            This has to be provided.
            This parameter is not used for mm2.0 metadata,
            which already provides positions

        Returns
        -------
        coord_to_filename   (dict)
            dictionary mapping of (key, value) = (coordinate, filepath)
            where coordinate = (position, time, channel, z-slice)
        """

        coords = set()
        meta = dict()

        # separate coords from meta
        for element in json_.keys():
            # present for mm2-gamma metadata
            if "Coords" in element:
                coords.add(element)
            if "Metadata" in element:
                meta["-".join(element.split("-")[1:])] = element

            # present in mm1.4.22 metadata
            if "FrameKey" in element:
                coords.add(element)

        if not coords:
            raise ValueError("no image coordinates present in metadata")

        # build a dict of coord to filename maps
        coord_to_filename = dict()
        for c in coords:
            # indices common to both mm2 and mm1
            ch_idx = json_[c]["ChannelIndex"]
            pos_idx = json_[c]["PositionIndex"]
            time_idx = json_[c]["FrameIndex"]
            z_idx = json_[c]["SliceIndex"]

            # extract filepath for this coordinate
            try:
                # for mm2-gamma. 'FileName' key contains position folder
                if "-".join(c.split("-")[1:]) in meta:
                    filepath = json_[meta["-".join(c.split("-")[1:])]][
                        "FileName"
                    ]
                # for mm1, 'FileName' key does not contain position folder
                else:
                    filepath = json_[c]["FileName"]
                    filepath = os.path.join(
                        position, filepath
                    )  # position name is explicitly supplied
            except KeyError as ke:
                self.log.error(
                    f"metadata for supplied image coordinate {c} not found"
                )
                raise ke

            coordinate = (pos_idx, time_idx, ch_idx, z_idx)
            coord_to_filename[coordinate] = os.path.join(
                parent_folder, filepath
            )

        return coord_to_filename

    def _get_sub_dirs(self, f):
        """
        subdir walk
        from https://github.com/mehta-lab/reconstruct-order

        Parameters
        ----------
        f:              (str)

        Returns
        -------
        sub_dir_name    (list) natsorted list of subdirectories
        """

        sub_dir_path = glob.glob(os.path.join(f, "*/"))
        sub_dir_name = [
            os.path.split(subdir[:-1])[1] for subdir in sub_dir_path
        ]
        #    assert subDirName, 'No sub directories found'
        return natsort.natsorted(sub_dir_name)

    def _dims_from_coordinates(self):
        """
        read coordinates from self.coord_to_filename
        parse the coordinates for the total number unique elements
        in each tuple position
        this total number reflects the true dimensionality
        coord = (pos, time, chan, z)
        height and width are still read from mm metadata
        Returns
        -------

        """
        p, t, c, z = set(), set(), set(), set()
        for coord in self.coord_to_filename.keys():
            p.add(coord[0])
            t.add(coord[1])
            c.add(coord[2])
            z.add(coord[3])
        self.frames = len(t)
        self.slices = len(z)
        self.channels = len(c)

    def _mm1_meta_parser(self):
        """
        set image metadata.
        from https://github.com/mehta-lab/reconstruct-order

        Returns
        -------

        """
        self.z_step_size = self._mm_meta["Summary"]["z-step_um"]
        self.width = self._mm_meta["Summary"]["Width"]
        self.height = self._mm_meta["Summary"]["Height"]
        self.frames = self._mm_meta["Summary"]["Frames"]
        self.slices = self._mm_meta["Summary"]["Slices"]
        self.channels = self._mm_meta["Summary"]["Channels"]

    def _mm2beta_meta_parser(self):
        """
        set image metadata
        from https://github.com/mehta-lab/reconstruct-order

        Returns
        -------

        """
        self.z_step_size = self._mm_meta["Summary"]["z-step_um"]
        self.width = int(
            self._mm_meta["Summary"]["UserData"]["Width"]["PropVal"]
        )
        self.height = int(
            self._mm_meta["Summary"]["UserData"]["Height"]["PropVal"]
        )
        self.time_stamp = self._mm_meta["Summary"]["StartTime"]

    def _mm2gamma_meta_parser(self):
        """
        set image metadata
        from https://github.com/mehta-lab/reconstruct-order

        Returns
        -------

        """
        keys_list = list(self._mm_meta.keys())
        if "FrameKey-0-0-0" in keys_list[1]:
            roi_string = self._mm_meta[keys_list[1]]["ROI"]
            self.width = int(roi_string.split("-")[2])
            self.height = int(roi_string.split("-")[3])
        elif "Metadata-" in keys_list[2]:
            self.width = self._mm_meta[keys_list[2]]["Width"]
            self.height = self._mm_meta[keys_list[2]]["Height"]
        else:
            raise ValueError("Metadata file incompatible with metadata reader")

        self.z_step_size = self._mm_meta["Summary"]["z-step_um"]
        self.frames = self._mm_meta["Summary"]["Frames"]
        self.slices = self._mm_meta["Summary"]["Slices"]
        self.channels = self._mm_meta["Summary"]["Channels"]



from __future__ import annotations

import os
import platform
import shutil
import string
from contextlib import contextmanager
from itertools import product
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Literal

import hypothesis.extra.numpy as npst
import hypothesis.strategies as st
import numpy as np
import ome_zarr.io
import ome_zarr.reader
import pytest
import zarr.storage
from hypothesis import HealthCheck, assume, given, settings
from ngff_zarr import from_ngff_zarr
from numpy.testing import assert_allclose, assert_array_equal
from numpy.typing import NDArray

if TYPE_CHECKING:
    from _typeshed import StrPath

from iohub.ngff.nodes import (
    TO_DICT_SETTINGS,
    Plate,
    Position,
    TransformationMeta,
    _case_insensitive_local_fs,
    _open_store,
    _pad_shape,
    open_ome_zarr,
)
from tests.conftest import hcs_ref

short_text_st = st.text(min_size=1, max_size=16)
t_dim_st = st.integers(1, 4)
c_dim_st = st.integers(1, 4)
z_dim_st = st.integers(1, 4)
y_dim_st = st.integers(1, 32)
x_dim_st = st.integers(1, 32)
channel_names_st = c_dim_st.flatmap(
    (
        lambda c_dim: st.lists(
            short_text_st, min_size=c_dim, max_size=c_dim, unique=True
        )
    )
)
ngff_versions_st = st.sampled_from(["0.4", "0.5"])
short_alpha_numeric = st.text(
    alphabet=list(
        string.ascii_lowercase + string.ascii_uppercase + string.digits
    ),
    min_size=1,
    max_size=16,
)
if os.name == "nt":
    # Windows does not allow certain file names
    _INVALID_NT_FILE_NAMES = (
        ["CON", "PRN", "AUX", "NUL"]
        + ["COM" + str(i) for i in range(10)]
        + ["LPT" + str(i) for i in range(10)]
    )
    short_alpha_numeric = short_alpha_numeric.filter(
        lambda x: x not in _INVALID_NT_FILE_NAMES
    )
tiles_rc_st = st.tuples(t_dim_st, t_dim_st)
plate_axis_names_st = st.lists(
    short_alpha_numeric,
    min_size=1,
    max_size=8,
    unique_by=(lambda x: x.lower()),
)


@st.composite
def _random_array_shape_and_dtype_with_channels(draw, c_dim: int):
    shape = (
        draw(t_dim_st),
        c_dim,
        draw(z_dim_st),
        draw(y_dim_st),
        draw(x_dim_st),
    )
    # zarr-python 3 broke big-endian support:
    # https://github.com/zarr-developers/zarr-python/issues/3005
    dtype = draw(
        st.one_of(
            npst.integer_dtypes(endianness="<"),
            npst.unsigned_integer_dtypes(endianness="<"),
            npst.floating_dtypes(endianness="<"),
            npst.boolean_dtypes(),
        )
    )
    return shape, dtype


@st.composite
def _channels_and_random_5d_shape_and_dtype(draw):
    channel_names = draw(channel_names_st)
    shape, dtype = draw(
        _random_array_shape_and_dtype_with_channels(c_dim=len(channel_names))
    )
    return channel_names, shape, dtype


@st.composite
def _channels_and_random_5d(draw):
    channel_names, shape, dtype = draw(
        _channels_and_random_5d_shape_and_dtype()
    )
    random_5d = draw(npst.arrays(dtype, shape=shape))
    return channel_names, random_5d


@given(shape=st.lists(x_dim_st, min_size=1, max_size=10), target=x_dim_st)
@settings(max_examples=16, deadline=1000)
def test_pad_shape(shape, target):
    """Test `iohub.ngff._pad_shape()`"""
    shape = tuple(shape)
    assume(len(shape) <= target)
    new_shape = _pad_shape(shape=shape, target=target)
    assert len(new_shape) == target
    assert new_shape[-len(shape) :] == shape


@given(version=ngff_versions_st)
def test_open_store_create(version):
    """Test `iohub.ngff._open_store()"""
    for mode in ("a", "w", "w-"):
        with TemporaryDirectory() as temp_dir:
            store_path = os.path.join(temp_dir, "new.zarr")
            root = _open_store(store_path, mode=mode, version=version)
            assert isinstance(root, zarr.Group)
            assert isinstance(root.store, zarr.storage.LocalStore)
            # assert root.store._dimension_separator == "/"
            assert root.store.root.resolve() == Path(store_path).resolve()


@given(version=ngff_versions_st)
def test_open_store_create_existing(version):
    """Test `iohub.ngff._open_store()"""
    with TemporaryDirectory() as temp_dir:
        store_path = os.path.join(temp_dir, "new.zarr")
        g = zarr.open_group(store_path, mode="w-")
        g.store.close()
        with pytest.raises(RuntimeError):
            _ = _open_store(store_path, mode="w-", version=version)
        assert _open_store(store_path, mode="w", version=version) is not None


@given(version=ngff_versions_st)
def test_open_store_read_nonexist(version):
    """Test `iohub.ngff._open_store()"""
    for mode in ("r", "r+"):
        with TemporaryDirectory() as temp_dir:
            store_path = os.path.join(temp_dir, "new.zarr")
            with pytest.raises(FileNotFoundError):
                _ = _open_store(store_path, mode=mode, version=version)


def test_case_insensitive_local_fs():
    """Test `iohub.ngff._case_insensitive_local_fs()`"""
    match platform.system():
        case "Windows":
            assert _case_insensitive_local_fs() is True
        case "Darwin":
            assert _case_insensitive_local_fs() is True
        case "Linux":
            assert _case_insensitive_local_fs() is False
        case _:
            _ = _case_insensitive_local_fs()


@given(channel_names=channel_names_st, version=ngff_versions_st)
@settings(max_examples=16)
def test_init_ome_zarr(channel_names, version):
    """Test `iohub.ngff.open_ome_zarr()`"""
    with TemporaryDirectory() as temp_dir:
        store_path = os.path.join(temp_dir, "ome.zarr")
        dataset = open_ome_zarr(
            store_path,
            layout="fov",
            mode="w-",
            channel_names=channel_names,
            version=version,
        )
        assert os.path.isdir(store_path)
        assert dataset.channel_names == channel_names


@pytest.mark.parametrize("version", ["0.4", "0.5"])
@pytest.mark.parametrize(
    "basename",
    ["some.zarr", "other.zarr/0/0/0", "random_dir", "napari_ome_zarr"],
)
def test_init_ome_zarr_overwrite_non_zarr(tmp_path, basename, version):
    """Test `iohub.ngff.open_ome_zarr()`"""
    store_path = tmp_path / basename
    store_path.mkdir(parents=True)
    some_child_directory = store_path / "some_other_directory"
    some_child_directory.mkdir()
    if ".zarr" not in basename:
        with pytest.raises(ValueError):
            _ = open_ome_zarr(
                store_path,
                layout="fov",
                mode="w",
                channel_names=["channel"],
                version=version,
            )
        assert some_child_directory.exists()
    assert (
        open_ome_zarr(
            store_path,
            layout="fov",
            mode="w",
            channel_names=["channel"],
            disable_path_checking=True,
            version=version,
        )
        is not None
    )
    assert not some_child_directory.exists()


@contextmanager
def _temp_ome_zarr(
    image_5d: NDArray,
    channel_names: list[str],
    arr_name: str,
    version: Literal["0.4", "0.5"],
    **kwargs,
):
    """Helper function to generate a temporary OME-Zarr store.

    Parameters
    ----------
    image_5d : NDArray
    channel_names : list[str]
    arr_name : str
    version : str
    **kwargs : dict
        Additional keyword arguments to pass to `create_image()`.

    Yields
    ------
    Position
    """
    try:
        temp_dir = TemporaryDirectory()
        dataset = open_ome_zarr(
            os.path.join(temp_dir.name, "ome.zarr"),
            layout="fov",
            mode="a",
            version=version,
            channel_names=channel_names,
        )
        dataset.create_image(arr_name, image_5d, **kwargs)
        yield dataset
    finally:
        dataset.close()
        temp_dir.cleanup()


@contextmanager
def _temp_ome_zarr_plate(
    image_5d: NDArray,
    channel_names: list[str],
    arr_name: str,
    position_list: list[tuple[str, str, str]],
    version: Literal["0.4", "0.5"],
    **kwargs,
):
    """Helper function to generate a temporary OME-Zarr store.

    Parameters
    ----------
    image_5d : NDArray
    channel_names : list[str]
    arr_name : str
    position_list : list[tuple[str, str, str]]
    version : Literal["0.4", "0.5"]

    Yields
    ------
    Position
    """
    try:
        temp_dir = TemporaryDirectory()
        dataset = open_ome_zarr(
            os.path.join(temp_dir.name, "ome.zarr"),
            layout="hcs",
            mode="a",
            channel_names=channel_names,
            version=version,
        )
        for position in position_list:
            pos = dataset.create_position(
                position[0], position[1], position[2]
            )
            pos.create_image(arr_name, image_5d, **kwargs)
        yield dataset
    finally:
        dataset.close()
        temp_dir.cleanup()


@given(
    channels_and_random_5d=_channels_and_random_5d(),
    arr_name=short_alpha_numeric,
    version=ngff_versions_st,
)
@settings(
    max_examples=16,
    deadline=2000,
    suppress_health_check=[HealthCheck.data_too_large],
)
def test_write_ome_zarr(channels_and_random_5d, arr_name, version):
    """Test `iohub.ngff.Position.__setitem__()`"""
    channel_names, random_5d = channels_and_random_5d
    with _temp_ome_zarr(
        random_5d, channel_names, arr_name, version=version
    ) as dataset:
        assert_allclose(dataset[arr_name][:], random_5d)
        # round-trip test with the offical reader implementation
        ext_reader = ome_zarr.reader.Reader(
            ome_zarr.io.parse_url(dataset.zgroup.store.root)
        )
        node = list(ext_reader())[0]
        assert node.metadata["channel_names"] == channel_names
        assert node.specs[0].datasets == [arr_name]
        assert_allclose(node.data[0], random_5d)


@given(
    ch_shape_dtype=_channels_and_random_5d_shape_and_dtype(),
    arr_name=short_alpha_numeric,
    version=ngff_versions_st,
)
@settings(
    max_examples=16,
    deadline=2000,
    suppress_health_check=[HealthCheck.data_too_large],
)
def test_create_zeros(ch_shape_dtype, arr_name, version):
    """Test `iohub.ngff.Position.create_zeros()`"""
    channel_names, shape, dtype = ch_shape_dtype
    with TemporaryDirectory() as temp_dir:
        store_path = os.path.join(temp_dir, "ome.zarr")
        dataset = open_ome_zarr(
            store_path,
            layout="fov",
            mode="w-",
            channel_names=channel_names,
            version=version,
        )
        dataset.create_zeros(name=arr_name, shape=shape, dtype=dtype)
        if version == "0.4":
            assert set(os.listdir(os.path.join(store_path, arr_name))) == {
                ".zarray",
                ".zattrs",
            }
        else:
            assert set(os.listdir(os.path.join(store_path, arr_name))) == {
                "zarr.json",
            }
            assert dataset[arr_name].metadata.dimension_names == (
                "T",
                "C",
                "Z",
                "Y",
                "X",
            )
        assert not dataset[arr_name][:].any()
        assert dataset[arr_name].shape == shape
        assert dataset[arr_name].dtype == dtype


@given(
    channels_and_random_5d=_channels_and_random_5d(),
    arr_name=short_alpha_numeric,
    version=ngff_versions_st,
)
@settings(
    max_examples=16,
    suppress_health_check=[HealthCheck.data_too_large],
)
def test_ome_zarr_to_dask(channels_and_random_5d, arr_name, version):
    """Test `iohub.ngff.Position.data` to dask"""
    channel_names, random_5d = channels_and_random_5d
    with _temp_ome_zarr(
        random_5d, channel_names, "0", version=version
    ) as dataset:
        assert_allclose(dataset.data.dask_array().compute(), random_5d)
    with _temp_ome_zarr(
        random_5d, channel_names, arr_name, version=version
    ) as dataset:
        assert_allclose(dataset[arr_name].dask_array().compute(), random_5d)


@given(channels_and_random_5d=_channels_and_random_5d())
@settings(
    max_examples=16,
    deadline=4000,
    suppress_health_check=[HealthCheck.data_too_large],
)
def test_writing_sharded(channels_and_random_5d):
    """Test `iohub.ngff.Position.data`"""
    channel_names, random_5d = channels_and_random_5d
    chunks = (
        1,
        max(1, random_5d.shape[1] // 3),
        max(1, random_5d.shape[2] // 4),
        max(1, random_5d.shape[3] // 5),
        max(1, random_5d.shape[4] // 6),
    )
    shards_ratio = (3, 4, 5, 6, 7)
    with _temp_ome_zarr(
        random_5d,
        channel_names,
        arr_name="0",
        version="0.5",
        chunks=chunks,
        shards_ratio=shards_ratio,
    ) as dataset:
        assert_array_equal(dataset["0"].numpy(), random_5d)
        assert dataset["0"].chunks == chunks
        assert dataset["0"].shards == tuple(
            c * s for c, s in zip(chunks, shards_ratio)
        )


@given(
    channels_and_random_5d=_channels_and_random_5d(),
    arr_name=short_alpha_numeric,
    version=ngff_versions_st,
)
@settings(
    max_examples=16,
    deadline=2000,
    suppress_health_check=[HealthCheck.data_too_large],
)
def test_position_data(channels_and_random_5d, arr_name, version):
    """Test `iohub.ngff.Position.data`"""
    channel_names, random_5d = channels_and_random_5d
    assume(arr_name != "0")
    with _temp_ome_zarr(
        random_5d, channel_names, "0", version=version
    ) as dataset:
        assert_allclose(dataset.data.numpy(), random_5d)
    with pytest.raises(KeyError):
        with _temp_ome_zarr(
            random_5d, channel_names, arr_name, version=version
        ) as dataset:
            _ = dataset.data


@given(
    channels_and_random_5d=_channels_and_random_5d(),
    arr_name=short_alpha_numeric,
    version=ngff_versions_st,
    concurrency=st.one_of(st.just(None), st.integers(1, 2)),
)
@settings(
    max_examples=16,
    deadline=2000,
    suppress_health_check=[HealthCheck.data_too_large],
)
def test_ome_zarr_to_tensorstore(
    channels_and_random_5d, arr_name, version, concurrency
):
    """Test `iohub.ngff.Position.data` to tensorstore"""
    import tensorstore as ts

    channel_names, random_5d = channels_and_random_5d
    with _temp_ome_zarr(
        random_5d, channel_names, arr_name, version=version
    ) as dataset:
        tstore = dataset[arr_name].tensorstore(
            context=(
                ts.Context({"data_copy_concurrency": {"limit": concurrency}})
                if concurrency is not None
                else None
            )
        )
        assert_array_equal(tstore, random_5d)
        zeros = np.zeros_like(random_5d)
        tstore[...].write(zeros).result()
        with open_ome_zarr(
            dataset.zgroup.store.root, mode="r"
        ) as read_only_dataset:
            assert_array_equal(read_only_dataset[arr_name].numpy(), zeros)
            read_only_tstore = read_only_dataset[arr_name].tensorstore()
            with pytest.raises(ValueError):
                read_only_tstore[0] = 1


@given(
    channels_and_random_5d=_channels_and_random_5d(),
    arr_name=short_alpha_numeric,
    version=ngff_versions_st,
)
@settings(
    max_examples=16,
    deadline=2000,
    suppress_health_check=[HealthCheck.data_too_large],
)
def test_append_channel(channels_and_random_5d, arr_name, version):
    """Test `iohub.ngff.Position.append_channel()`"""
    channel_names, random_5d = channels_and_random_5d
    assume(len(channel_names) > 1)
    with _temp_ome_zarr(
        random_5d[:, :-1], channel_names[:-1], arr_name, version=version
    ) as dataset:
        dataset.append_channel(channel_names[-1], resize_arrays=True)
        dataset[arr_name][:, -1] = random_5d[:, -1]
        assert_allclose(dataset[arr_name][:], random_5d)


@given(
    channels_and_random_5d=_channels_and_random_5d(),
    arr_name=short_alpha_numeric,
    version=ngff_versions_st,
    new_channel=short_text_st,
)
@settings(
    max_examples=16,
    deadline=2000,
    suppress_health_check=[HealthCheck.data_too_large],
)
def test_rename_channel(
    channels_and_random_5d, arr_name, new_channel, version
):
    """Test `iohub.ngff.Position.rename_channel()`"""
    channel_names, random_5d = channels_and_random_5d
    assume(new_channel not in channel_names)
    with _temp_ome_zarr(
        random_5d, channel_names, arr_name, version=version
    ) as dataset:
        dataset.rename_channel(old=channel_names[0], new=new_channel)
        assert new_channel in dataset.channel_names
        assert dataset.metadata.omero.channels[0].label == new_channel


@given(
    channels_and_random_5d=_channels_and_random_5d(),
    arr_name=short_alpha_numeric,
    version=ngff_versions_st,
)
@settings(deadline=None)
def test_rename_well(channels_and_random_5d, arr_name, version):
    """Test `iohub.ngff.Position.rename_well()`"""
    channel_names, random_5d = channels_and_random_5d

    position_list = [("A", "1", "0"), ("C", "4", "0")]
    with _temp_ome_zarr_plate(
        random_5d, channel_names, arr_name, position_list, version
    ) as dataset:
        assert dataset.zgroup["A/1"]
        with pytest.raises(KeyError):
            dataset.zgroup["B/2"]
        assert "A" in [r[0] for r in dataset.rows()]
        assert "B" not in [r[0] for r in dataset.rows()]
        assert "A" in [row.name for row in dataset.metadata.rows]
        assert "B" not in [row.name for row in dataset.metadata.rows]
        assert "1" in [col.name for col in dataset.metadata.columns]
        assert "2" not in [col.name for col in dataset.metadata.columns]
        assert "C" in [row.name for row in dataset.metadata.rows]
        assert "4" in [col.name for col in dataset.metadata.columns]

        dataset.rename_well("A/1", "B/2")

        assert dataset.zgroup["B/2"]
        with pytest.raises(KeyError):
            dataset.zgroup["A/1"]
        assert "A" not in [r[0] for r in dataset.rows()]
        assert "B" in [r[0] for r in dataset.rows()]
        assert "A" not in [row.name for row in dataset.metadata.rows]
        assert "B" in [row.name for row in dataset.metadata.rows]
        assert "1" not in [col.name for col in dataset.metadata.columns]
        assert "2" in [col.name for col in dataset.metadata.columns]
        assert "C" in [row.name for row in dataset.metadata.rows]
        assert "4" in [col.name for col in dataset.metadata.columns]

        # destination exists
        with pytest.raises(FileExistsError):
            dataset.rename_well("B/2", "C/4")

        # source doesn't exist
        with pytest.raises(FileNotFoundError):
            dataset.rename_well("Q/1", "Q/2")

        # invalid well names
        with pytest.raises(ValueError):
            dataset.rename_well("B/2", " A/1")
        with pytest.raises(ValueError):
            dataset.rename_well("B/2", "A/?")


@given(
    channels_and_random_5d=_channels_and_random_5d(),
    arr_name=short_alpha_numeric,
    version=ngff_versions_st,
)
@settings(
    max_examples=16,
    deadline=2000,
    suppress_health_check=[HealthCheck.data_too_large],
)
def test_update_channel(channels_and_random_5d, arr_name, version):
    """Test `iohub.ngff.Position.update_channel()`"""
    channel_names, random_5d = channels_and_random_5d
    assume(len(channel_names) > 1)
    with _temp_ome_zarr(
        random_5d[:, :-1], channel_names[:-1], arr_name, version=version
    ) as dataset:
        for i, ch in enumerate(dataset.channel_names):
            dataset.update_channel(
                chan_name=ch, target=arr_name, data=random_5d[:, -1]
            )
            assert_allclose(dataset[arr_name][:, i], random_5d[:, -1])


@given(
    channels_and_random_5d=_channels_and_random_5d(),
    arr_name=short_alpha_numeric,
    version=ngff_versions_st,
)
@settings(
    max_examples=16,
    deadline=2000,
    suppress_health_check=[HealthCheck.data_too_large],
)
def test_write_more_channels(channels_and_random_5d, arr_name, version):
    """Test `iohub.ngff.Position.create_image()`"""
    channel_names, random_5d = channels_and_random_5d
    assume(len(channel_names) > 1)
    with pytest.raises(ValueError):
        with _temp_ome_zarr(
            random_5d, channel_names[:-1], arr_name, version=version
        ) as _:
            pass


@given(
    ch_shape_dtype=_channels_and_random_5d_shape_and_dtype(),
    arr_name=short_alpha_numeric,
)
def test_set_transform_image(ch_shape_dtype, arr_name):
    """Test `iohub.ngff.Position.set_transform()`"""
    channel_names, shape, dtype = ch_shape_dtype
    transform = [
        TransformationMeta(type="translation", translation=(1, 2, 3, 4, 5))
    ] * len(channel_names)
    with TemporaryDirectory() as temp_dir:
        store_path = os.path.join(temp_dir, "ome.zarr")
        with open_ome_zarr(
            store_path, layout="fov", mode="w-", channel_names=channel_names
        ) as dataset:
            dataset.create_zeros(name=arr_name, shape=shape, dtype=dtype)
            assert dataset.metadata.multiscales[0].datasets[
                0
            ].coordinate_transformations == [
                TransformationMeta(
                    type="scale", scale=(1.0, 1.0, 1.0, 1.0, 1.0)
                )
            ]
            dataset.set_transform(image=arr_name, transform=transform)
            assert (
                dataset.metadata.multiscales[0]
                .datasets[0]
                .coordinate_transformations
                == transform
            )
        # read data with an external reader
        ext_reader = ome_zarr.reader.Reader(
            ome_zarr.io.parse_url(dataset.zgroup.store.root)
        )
        node = list(ext_reader())[0]
        assert node.metadata["coordinateTransformations"][0] == [
            translate.model_dump(**TO_DICT_SETTINGS) for translate in transform
        ]


input_transformations = [
    ([TransformationMeta(type="identity")], []),
    ([TransformationMeta(type="scale", scale=(1.0, 2.0, 3.0, 4.0, 5.0))], []),
    (
        [
            TransformationMeta(
                type="translation", translation=(1.0, 2.0, 3.0, 4.0, 5.0)
            )
        ],
        [],
    ),
    (
        [
            TransformationMeta(type="scale", scale=(2.0, 2.0, 2.0, 2.0, 2.0)),
            TransformationMeta(
                type="translation", translation=(1.0, 1.0, 1.0, 1.0, 1.0)
            ),
        ],
        [
            TransformationMeta(type="scale", scale=(2.0, 2.0, 2.0, 2.0, 2.0)),
            TransformationMeta(
                type="translation", translation=(1.0, 1.0, 1.0, 1.0, 1.0)
            ),
        ],
    ),
]
target_scales = [
    [1.0, 1.0, 1.0, 1.0, 1.0],
    [1.0, 2.0, 3.0, 4.0, 5.0],
    [1.0, 1.0, 1.0, 1.0, 1.0],
    [4.0, 4.0, 4.0, 4.0, 4.0],
]
target_translations = [
    [0.0, 0.0, 0.0, 0.0, 0.0],
    [0.0, 0.0, 0.0, 0.0, 0.0],
    [1.0, 2.0, 3.0, 4.0, 5.0],
    [2.0, 2.0, 2.0, 2.0, 2.0],
]


@pytest.mark.parametrize(
    "transforms",
    [
        (saved, target)
        for saved, target in zip(input_transformations, target_scales)
    ],
)
@given(
    ch_shape_dtype=_channels_and_random_5d_shape_and_dtype(),
    arr_name=short_alpha_numeric,
    version=ngff_versions_st,
)
def test_get_effective_scale_image(
    transforms, ch_shape_dtype, arr_name, version
):
    """Test `iohub.ngff.Position.get_effective_scale()`"""
    (fov_transform, img_transform), expected_scale = transforms
    channel_names, shape, dtype = ch_shape_dtype
    with TemporaryDirectory() as temp_dir:
        store_path = os.path.join(temp_dir, "ome.zarr")
        with open_ome_zarr(
            store_path,
            layout="fov",
            mode="w-",
            channel_names=channel_names,
            version=version,
        ) as dataset:
            dataset.create_zeros(name=arr_name, shape=shape, dtype=dtype)
            dataset.set_transform(image="*", transform=fov_transform)
            dataset.set_transform(image=arr_name, transform=img_transform)
            scale = dataset.get_effective_scale(image=arr_name)
            assert scale == expected_scale


@pytest.mark.parametrize(
    "transforms",
    [
        (saved, target)
        for saved, target in zip(input_transformations, target_translations)
    ],
)
@given(
    ch_shape_dtype=_channels_and_random_5d_shape_and_dtype(),
    arr_name=short_alpha_numeric,
    version=ngff_versions_st,
)
def test_get_effective_translation_image(
    transforms, ch_shape_dtype, arr_name, version
):
    """Test `iohub.ngff.Position.get_effective_translation()`"""
    (fov_transform, img_transform), expected_translation = transforms
    channel_names, shape, dtype = ch_shape_dtype
    with TemporaryDirectory() as temp_dir:
        store_path = os.path.join(temp_dir, "ome.zarr")
        with open_ome_zarr(
            store_path,
            layout="fov",
            mode="w-",
            channel_names=channel_names,
            version=version,
        ) as dataset:
            dataset.create_zeros(name=arr_name, shape=shape, dtype=dtype)
            dataset.set_transform(image="*", transform=fov_transform)
            dataset.set_transform(image=arr_name, transform=img_transform)
            translation = dataset.get_effective_translation(image=arr_name)
            assert translation == expected_translation


@given(
    ch_shape_dtype=_channels_and_random_5d_shape_and_dtype(),
    arr_name=short_alpha_numeric,
    version=ngff_versions_st,
)
def test_set_transform_fov(ch_shape_dtype, arr_name, version):
    """Test `iohub.ngff.Position.set_transform()`"""
    channel_names, shape, dtype = ch_shape_dtype
    transform = [
        TransformationMeta(type="translation", translation=(1, 2, 3, 4, 5))
    ] * len(channel_names)
    with TemporaryDirectory() as temp_dir:
        store_path = os.path.join(temp_dir, "ome.zarr")
        with open_ome_zarr(
            store_path,
            layout="fov",
            mode="w-",
            channel_names=channel_names,
            version=version,
        ) as dataset:
            dataset.create_zeros(name=arr_name, shape=shape, dtype=dtype)
            assert (
                dataset.metadata.multiscales[0].coordinate_transformations
                == None
            )
            dataset.set_transform(image="*", transform=transform)
            assert (
                dataset.metadata.multiscales[0].coordinate_transformations
                == transform
            )
        # read data with plain zarr
        group = zarr.open(store_path)
        if version == "0.4":
            maybe_ome = group.attrs
        elif version == "0.5":
            maybe_ome = group.attrs["ome"]
        assert maybe_ome["multiscales"][0]["coordinateTransformations"] == [
            translate.model_dump(**TO_DICT_SETTINGS) for translate in transform
        ]


@pytest.mark.parametrize("version", ["0.4", "0.5"])
@pytest.mark.parametrize("image_name", ["0", "1", "a", "*"])
def test_set_scale(image_name, version):
    """Test `iohub.ngff.Position.set_scale()`"""
    translation = [float(t) for t in range(1, 6)]
    scale = [float(s) for s in range(5, 0, -1)]
    array_name = "0" if image_name == "*" else image_name
    new_scale = 10.0
    with TemporaryDirectory() as temp_dir:
        store_path = os.path.join(temp_dir, "ome.zarr")
        with open_ome_zarr(
            store_path,
            layout="fov",
            mode="w-",
            channel_names=["a", "b"],
            version=version,
        ) as dataset:
            dataset.create_zeros(
                name=array_name,
                shape=(1, 2, 4, 8, 16),
                dtype=int,
                transform=[
                    TransformationMeta(
                        type="translation", translation=translation
                    ),
                    TransformationMeta(type="scale", scale=scale),
                ],
            )
            with pytest.raises(ValueError):
                dataset.set_scale(
                    image=image_name, axis_name="z", new_scale=-1.0
                )
            with pytest.raises(KeyError):
                dataset.set_scale(
                    image="nonexistent", axis_name="z", new_scale=9.0
                )
            assert dataset.scale[-3] == 3.0
            dataset.set_scale(
                image=image_name, axis_name="z", new_scale=new_scale
            )
            if image_name == "*":
                assert dataset.scale[-3] == new_scale * 3.0
            else:
                assert dataset.scale[-3] == new_scale
            assert dataset.get_effective_translation(array_name) == translation
            for tf in dataset.zattrs["iohub"]["previous_transforms"][0][
                "transforms"
            ]:
                if tf["type"] == "scale":
                    assert tf["scale"] == scale


@given(channel_names=channel_names_st, version=ngff_versions_st)
@settings(max_examples=16)
def test_set_contrast_limits(channel_names, version):
    """Test `iohub.ngff.Position.set_contrast_limits()`"""
    with TemporaryDirectory() as temp_dir:
        store_path = os.path.join(temp_dir, "ome.zarr")
        dataset = open_ome_zarr(
            store_path,
            layout="fov",
            mode="a",
            channel_names=channel_names,
            version=version,
        )
        # Create a simple small array - exact shape/dtype doesn't matter
        dataset.create_zeros(
            "data", shape=(1, len(channel_names), 1, 4, 4), dtype=float
        )

        # Store the initial window settings for all channels
        initial_windows = {}
        for ch_name in channel_names:
            ch_idx = dataset.get_channel_index(ch_name)
            initial_windows[ch_name] = dataset.metadata.omero.channels[
                ch_idx
            ].window

        # Test setting contrast limits for the first channel only
        target_channel = channel_names[0]
        window = {"start": 10.0, "end": 100.0, "min": 0.0, "max": 255.0}

        # Set contrast limits
        dataset.set_contrast_limits(target_channel, window)

        # Check that the contrast limits
        # were set correctly for the target channel
        channel_index = dataset.get_channel_index(target_channel)
        channel_window = dataset.metadata.omero.channels[channel_index].window
        assert channel_window is not None
        assert channel_window["start"] == window["start"]
        assert channel_window["end"] == window["end"]
        assert channel_window["min"] == window["min"]
        assert channel_window["max"] == window["max"]

        # Check that other channels were not affected (if any exist)
        for ch_name in channel_names[1:]:
            ch_idx = dataset.get_channel_index(ch_name)
            assert (
                dataset.metadata.omero.channels[ch_idx].window
                == initial_windows[ch_name]
            )


@given(channel_names=channel_names_st, version=ngff_versions_st)
@settings(max_examples=16)
def test_create_tiled(channel_names, version):
    """Test that `iohub.ngff.open_ome_zarr()` can create
    an empty OME-Zarr store with 'tiled' layout."""
    with TemporaryDirectory() as temp_dir:
        store_path = os.path.join(temp_dir, "tiled.zarr")
        dataset = open_ome_zarr(
            store_path,
            layout="tiled",
            mode="a",
            channel_names=channel_names,
            version=version,
        )
        assert os.path.isdir(store_path)
        assert dataset.channel_names == channel_names


@given(
    channels_and_random_5d=_channels_and_random_5d(),
    grid_shape=tiles_rc_st,
    arr_name=short_alpha_numeric,
)
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_make_tiles(channels_and_random_5d, grid_shape, arr_name):
    """Test `iohub.ngff.TiledPosition.make_tiles()` and  `...get_tile()`"""
    with TemporaryDirectory() as temp_dir:
        channel_names, random_5d = channels_and_random_5d
        store_path = os.path.join(temp_dir, "tiled.zarr")
        with open_ome_zarr(
            store_path, layout="tiled", mode="a", channel_names=channel_names
        ) as dataset:
            tiles = dataset.make_tiles(
                name=arr_name,
                grid_shape=(int(grid_shape[0]), int(grid_shape[1])),
                tile_shape=tuple(int(i) for i in random_5d.shape),
                dtype=random_5d.dtype,
                chunk_dims=2,
            )
            assert tiles.rows == grid_shape[0]
            assert tiles.columns == grid_shape[1]
            assert tiles.tiles == grid_shape
            assert tiles.shape[-2:] == (
                grid_shape[-2] * random_5d.shape[-2],
                grid_shape[-1] * random_5d.shape[-1],
            )
            assert tiles.tile_shape == _pad_shape(
                random_5d.shape[-2:], target=5
            )
            assert tiles.dtype == random_5d.dtype
            for args in ((1.01, 1), (0, 0, 0)):
                with pytest.raises(TypeError):
                    tiles.get_tile(*args)
            for args in ((0, 0, (0,) * 2), (0, 0, (0,) * 4)):
                with pytest.raises(IndexError):
                    tiles.get_tile(*args)


@given(
    channels_and_random_5d=_channels_and_random_5d(),
    grid_shape=tiles_rc_st,
    arr_name=short_alpha_numeric,
    version=ngff_versions_st,
)
@settings(
    max_examples=16,
    deadline=2000,
    suppress_health_check=[HealthCheck.data_too_large],
)
def test_write_read_tiles(
    channels_and_random_5d, grid_shape, arr_name, version
):
    """Test `iohub.ngff.TiledPosition.write_tile()` and `...get_tile()`"""
    channel_names, random_5d = channels_and_random_5d

    def _tile_data(tiles):
        for row in range(tiles.rows):
            for column in range(tiles.columns):
                yield (
                    (
                        random_5d
                        / (tiles.rows * tiles.columns + 1)
                        * (row * column + 1)
                    ).astype(random_5d.dtype),
                    row,
                    column,
                )

    with TemporaryDirectory() as temp_dir:
        store_path = os.path.join(temp_dir, "tiled.zarr")
        with open_ome_zarr(
            store_path,
            layout="tiled",
            mode="w-",
            channel_names=channel_names,
            version=version,
        ) as dataset:
            tiles = dataset.make_tiles(
                name=arr_name,
                grid_shape=grid_shape,
                tile_shape=random_5d.shape,
                dtype=random_5d.dtype,
                chunk_dims=2,
            )
            for data, row, column in _tile_data(tiles):
                tiles.write_tile(data, row, column)
        with open_ome_zarr(
            store_path, layout="tiled", mode="r", channel_names=channel_names
        ) as dataset:
            for data, row, column in _tile_data(tiles):
                read = tiles.get_tile(row, column)
                assert_allclose(data, read)


@given(channel_names=channel_names_st)
@settings(max_examples=16)
def test_create_hcs(channel_names):
    """Test `iohub.ngff.open_ome_zarr()`"""
    with TemporaryDirectory() as temp_dir:
        store_path = os.path.join(temp_dir, "hcs.zarr")
        dataset = open_ome_zarr(
            store_path, layout="hcs", mode="a", channel_names=channel_names
        )
        assert os.path.isdir(store_path)
        assert dataset.channel_names == channel_names


@pytest.mark.parametrize("version", ["0.4", "0.5"])
def test_open_hcs_create_empty(version):
    """Test `iohub.ngff.open_ome_zarr()`"""
    with TemporaryDirectory() as temp_dir:
        store_path = Path(temp_dir) / "hcs.zarr"
        dataset = open_ome_zarr(
            store_path,
            layout="hcs",
            mode="a",
            channel_names=["GFP"],
            version=version,
        )
        assert dataset.zgroup.store.root.resolve() == store_path.resolve()
        dataset.close()
        with pytest.raises(FileExistsError):
            _ = open_ome_zarr(
                store_path, layout="hcs", mode="w-", channel_names=["GFP"]
            )
        with pytest.raises(ValueError):
            _ = open_ome_zarr(store_path, layout="hcs", mode="x")
        with pytest.raises(FileNotFoundError):
            _ = open_ome_zarr("do-not-exist.zarr", layout="hcs", mode="r+")
        with pytest.raises(ValueError):
            dataset = open_ome_zarr(store_path, layout="hcs", mode="r+")


@contextmanager
def _temp_copy(src: StrPath):
    """Create a temporary copy of data on disk."""
    try:
        temp_dir = TemporaryDirectory()
        yield shutil.copytree(src, temp_dir.name, dirs_exist_ok=True)
    finally:
        temp_dir.cleanup()


@given(wrong_channel_name=channel_names_st)
def test_get_channel_index(wrong_channel_name):
    """Test `iohub.ngff.NGFFNode.get_channel_axis()`"""
    assume(wrong_channel_name != "DAPI")
    with open_ome_zarr(hcs_ref, layout="hcs", mode="r+") as dataset:
        assert dataset.get_channel_index("DAPI") == 0
        with pytest.raises(ValueError):
            _ = dataset.get_channel_index(wrong_channel_name)


def test_get_axis_index():
    with open_ome_zarr(hcs_ref, layout="hcs", mode="r+") as dataset:
        position = dataset["B/03/0"]

        assert position.axis_names == ["c", "z", "y", "x"]

        assert position.get_axis_index("z") == 1
        assert position.get_axis_index("Z") == 1

        with pytest.raises(ValueError):
            _ = position.get_axis_index("t")

        with pytest.raises(ValueError):
            _ = position.get_axis_index("DOG")


def test_ngff_node_contains_cross_platform(caplog):
    """Test `iohub.ngff.NGFFNode.__contains__()` on multiple platforms."""
    with open_ome_zarr(hcs_ref, layout="hcs", mode="r") as dataset:
        assert "B" in dataset
        match platform.system():
            case "Linux":
                assert "b" not in dataset
            case "Windows" | "Darwin":
                assert "b" in dataset
                assert any(
                    "Key 'b' matched" in r.message for r in caplog.records
                )


@given(
    row=short_alpha_numeric, col=short_alpha_numeric, pos=short_alpha_numeric
)
@settings(max_examples=16, deadline=2000)
def test_modify_hcs_ref(row: str, col: str, pos: str):
    """Test `iohub.ngff.open_ome_zarr()`"""
    assume((row.lower() != "b"))
    with _temp_copy(hcs_ref) as store_path:
        with open_ome_zarr(store_path, layout="hcs", mode="r+") as dataset:
            assert dataset.axes[0].name == "c"
            assert dataset.channel_names == ["DAPI"]
            position = dataset["B/03/0"]
            assert position[0].shape == (1, 2, 2160, 5120)
            position.append_channel("GFP", resize_arrays=True)
            assert position.channel_names == ["DAPI", "GFP"]
            assert position[0].shape == (2, 2, 2160, 5120)
            new_pos_path = "/".join([row, col, pos])
            assume(new_pos_path not in dataset)
            new_pos = dataset.create_position(row, col, pos)
            new_pos.create_zeros("0", position[0].shape, position[0].dtype)
            assert not dataset[f"{new_pos_path}/0"][:].any()


@given(row_names=plate_axis_names_st, col_names=plate_axis_names_st)
@settings(max_examples=16, deadline=2000)
def test_create_well(row_names: list[str], col_names: list[str]):
    """Test `iohub.ngff.Plate.create_well()`"""
    with TemporaryDirectory() as temp_dir:
        store_path = os.path.join(temp_dir, "hcs.zarr")
        dataset = open_ome_zarr(
            store_path, layout="hcs", mode="a", channel_names=["GFP"]
        )
        for row_name in row_names:
            for col_name in col_names:
                dataset.create_well(row_name, col_name)
        assert [
            c["name"] for c in dataset.zattrs["plate"]["columns"]
        ] == col_names
        assert [
            r["name"] for r in dataset.zattrs["plate"]["rows"]
        ] == row_names


def test_create_case_sensitive_well(tmp_path):
    """Test `iohub.ngff.Plate.create_well()` with case-sensitive names."""
    store_path = tmp_path / "hcs.zarr"
    with open_ome_zarr(
        store_path, layout="hcs", mode="w-", channel_names=["1", "2"]
    ) as dataset:
        well = dataset.create_well("A", "B")
        fov = well.create_position("0")
        fov.create_zeros("0", shape=(1, 2, 3, 4, 5), dtype=int)
        match platform.system():
            case "Windows" | "Darwin":
                with pytest.raises(FileExistsError):
                    dataset.create_well("a", "B")
                with pytest.raises(FileExistsError):
                    dataset.create_well("A", "b")
                new_well = dataset.create_well("a", "1")
                expected_rows = 1
            case "Linux":
                new_well = dataset.create_well("a", "b")
                expected_rows = 2
        new_fov = new_well.create_position("0")
        new_fov.create_zeros("0", shape=(1, 2, 3, 4, 5), dtype=int)
    with open_ome_zarr(store_path) as dataset:
        assert len(dataset.metadata.rows) == expected_rows
        assert len(list(dataset.rows())) == expected_rows
        assert len(dataset.metadata.columns) == 2


@given(
    row=short_alpha_numeric,
    col=short_alpha_numeric,
    pos=short_alpha_numeric,
    version=ngff_versions_st,
)
def test_create_position(row, col, pos, version):
    """Test `iohub.ngff.Plate.create_position()`"""
    with TemporaryDirectory() as temp_dir:
        store_path = Path(temp_dir) / "hcs.zarr"
        dataset = open_ome_zarr(
            store_path,
            layout="hcs",
            mode="a",
            channel_names=["GFP"],
            version=version,
        )
        _ = dataset.create_position(row_name=row, col_name=col, pos_name=pos)
        if version == "0.4":
            ome = dataset.zgroup.attrs
        elif version == "0.5":
            ome = dataset.zgroup.attrs["ome"]
        assert [c["name"] for c in ome["plate"]["columns"]] == [col]
        assert [r["name"] for r in ome["plate"]["rows"]] == [row]
        assert (store_path / row / col / pos).is_dir()
        assert dataset[row][col].metadata.images[0].path == pos


@given(
    channels_and_random_5d=_channels_and_random_5d(), version=ngff_versions_st
)
def test_position_scale(channels_and_random_5d, version):
    """Test `iohub.ngff.Position.scale`"""
    channel_names, random_5d = channels_and_random_5d
    scale = list(range(1, 6))
    transform = [TransformationMeta(type="scale", scale=scale)]
    with _temp_ome_zarr(
        random_5d, channel_names, "0", transform=transform, version=version
    ) as dataset:
        assert dataset.scale == scale


@pytest.mark.skip(
    reason="https://github.com/zarr-developers/zarr-python/issues/2407"
)
def test_combine_fovs_to_hcs():
    fovs = {}
    fov_paths = ("A/1/0", "B/1/0", "H/12/9")
    with open_ome_zarr(hcs_ref) as hcs_store:
        fov = hcs_store["B/03/0"]
        array = fov[0].numpy()
        channel_names = fov.channel_names
        old_omero_name = fov.metadata.omero.name
        for path in fov_paths:
            fovs[path] = fov
    with TemporaryDirectory() as temp_dir:
        store_path = os.path.join(temp_dir, "combined.zarr")
        Plate.from_positions(store_path, fovs).close()
        with open_ome_zarr(store_path, layout="hcs", mode="r") as dataset:
            assert len(dataset.metadata.rows) == 3
            assert len(dataset.metadata.columns) == 2
            for fov_path in fov_paths:
                copied_fov = dataset[fov_path]
                assert copied_fov.channel_names == channel_names
                new_omero_name = copied_fov.metadata.omero.name
                assert new_omero_name != old_omero_name
                assert new_omero_name == fov_path[-1]
                assert_array_equal(dataset[fov_path]["0"].numpy(), array)


def test_hcs_external_reader(tmp_path):
    store_path = tmp_path / "hcs.zarr"
    fov_name_parts = (("A", "1", "7"), ("B", "1", "7"), ("H", "12", "7"))
    y_size, x_size = (128, 100)
    with open_ome_zarr(
        store_path, layout="hcs", mode="a", channel_names=["A", "B"]
    ) as dataset:
        for name in fov_name_parts:
            fov = dataset.create_position(*name)
            fov.create_zeros("0", shape=(1, 2, 3, y_size, x_size), dtype=int)
        n_rows = len(dataset.metadata.rows)
        n_cols = len(dataset.metadata.columns)
    plate = list(ome_zarr.reader.Reader(ome_zarr.io.parse_url(store_path))())[
        0
    ]
    assert plate.data[0].shape == (1, 2, 3, y_size * n_rows, x_size * n_cols)
    assert plate.data[0].dtype == int
    assert not plate.data[0].any()
    assert plate.metadata["channel_names"] == ["A", "B"]


def test_read_empty_hcs_v05(empty_ome_zarr_hcs_v05):
    """Test reading an empty OME-Zarr v0.5 HCS store."""
    empty_zarr, (rows, cols, fovs, resolutions) = empty_ome_zarr_hcs_v05
    with open_ome_zarr(empty_zarr, layout="hcs", mode="r") as dataset:
        for row, col, fov in product(rows, cols, fovs):
            position: Position = dataset[f"{row}/{col}/{fov}"]
            assert position.version == "0.5"
            for resolution in resolutions:
                assert_array_equal(
                    position[resolution].numpy(),
                    np.zeros((50, 48, 64), dtype=np.uint16),
                )
        assert len(list(dataset.positions())) == len(rows) * len(cols) * len(
            fovs
        )


def test_acquire_zarr_ome_zarr_05(aqz_ome_zarr_05):
    """Test that `iohub.ngff.open_ome_zarr()` can read OME-Zarr 0.5."""
    pytest.importorskip("acquire_zarr")
    with open_ome_zarr(
        aqz_ome_zarr_05, layout="fov", mode="r", version="0.5"
    ) as dataset:
        assert dataset.version == "0.5"
        assert dataset.data.shape == (32, 4, 10, 48, 64)
        assert dataset.data.chunks == (16, 1, 10, 16, 16)
        assert dataset.data.shards == (16, 1, 10, 48, 32)
        assert "ome" in dataset.zattrs
        assert "multiscales" in dataset.zattrs["ome"]
        assert len(dataset.zattrs["ome"]["multiscales"]) == 1

        multiscale = dataset.zattrs["ome"]["multiscales"][0]
        assert len(multiscale["datasets"]) == 3
        assert multiscale["datasets"][0]["coordinateTransformations"][0][
            "scale"
        ] == [1.0, 1.0, 1.0, 1.0, 1.0]
        assert multiscale["datasets"][1]["coordinateTransformations"][0][
            "scale"
        ] == [1.0, 1.0, 1.0, 2.0, 2.0]
        assert multiscale["datasets"][2]["coordinateTransformations"][0][
            "scale"
        ] == [1.0, 1.0, 1.0, 4.0, 4.0]
        assert 1 < dataset["0"].numpy().mean() < np.iinfo(np.uint16).max


@given(
    channels_and_random_5d=_channels_and_random_5d(),
    arr_name=short_alpha_numeric,
    version=ngff_versions_st,
)
@settings(
    max_examples=16,
    suppress_health_check=[HealthCheck.data_too_large],
)
def test_ngff_zarr_read(channels_and_random_5d, arr_name, version):
    """Test that image written with iohub can be read with ngff-zarr."""
    channel_names, random_5d = channels_and_random_5d
    with _temp_ome_zarr(
        random_5d, channel_names, arr_name=arr_name, version=version
    ) as dataset:
        nz_multiscales = from_ngff_zarr(
            dataset.zgroup.store.root, validate=True
        )
        assert_allclose(
            dataset[arr_name].dask_array().compute(),
            nz_multiscales.images[0].data,
        )



"""
Update OME-Zarr Version
=======================

This script shows how to write the same OME-Zarr image
using a new version.
"""

# %%
from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np

from iohub.ngff import TransformationMeta, open_ome_zarr

# %%
# Set storage path
tmp_dir = TemporaryDirectory()
old_store_path = Path(tmp_dir.name) / "old.zarr"
new_store_path = Path(tmp_dir.name) / "new.zarr"

# %%
# Create a version 0.4 OME-Zarr dataset
random_image = np.random.randint(
    0, np.iinfo(np.uint16).max, size=(10, 2, 32, 128, 128), dtype=np.uint16
)
scale = [2.0, 3.0, 4.0, 5.0, 6.0]


with open_ome_zarr(
    old_store_path,
    layout="hcs",
    mode="w-",
    channel_names=["DAPI", "GFP"],
    version="0.4",
) as old_dataset:
    position = old_dataset.create_position("A", "1", "0")
    image = position.create_image(
        "0",
        random_image,
        chunks=(1, 1, 4, 32, 32),
        transform=[TransformationMeta(type="scale", scale=scale)],
    )

# %%
# Write the same image with version 0.5 and sharding

with open_ome_zarr(old_store_path, mode="r", layout="hcs") as old_dataset:
    with open_ome_zarr(
        new_store_path,
        layout="hcs",
        mode="w",
        channel_names=old_dataset.channel_names,
        version="0.5",
    ) as new_dataset:
        for name, old_position in old_dataset.positions():
            row, col, fov = name.split("/")
            new_position = new_dataset.create_position(row, col, fov)
            old_image = old_position["0"]
            new_image = new_position.create_image(
                "0",
                data=old_image.numpy(),
                chunks=(1, 1, 4, 32, 32),
                shards_ratio=(2, 1, 8, 4, 4),
                transform=old_position.metadata.multiscales[0]
                .datasets[0]
                .coordinate_transformations,
            )

# %%
# Read the new FOV to verify it was written correctly
with open_ome_zarr(new_store_path / "A/1/0", mode="r") as dataset:
    assert dataset.scale == scale
    image = dataset["0"]
    assert image.shards == (2, 1, 32, 128, 128)
    assert np.array_equal(image.numpy(), random_image)

# %%
# Clean up
tmp_dir.cleanup()
